/*
	Copyright (C) 2008, 2009 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef MULTITHREADED
#include <pthread.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif

#include "common.h"
#include "algo.h"
#include "util.h"

/*!
 * \brief Allocates a two-dimensional \p size x \p size
 * array of UINT pointers used for storing bucket boundaries.
 *
 * \param size Width and height of the array / number of buckets.
 * \returns Pointer to the array.
 */
static UINT*** alloc_buckets(const size_t size) {
	UINT*** array = malloc_safe((size * sizeof(UINT**)) +
			(size * size * sizeof(UINT*)));
	assert(array != NULL);

	size_t i;
	UINT** offset = (UINT**) array + size;

	for(i = 0; i < size; i++) {
		*(array + i) = offset + (i * size); 
	}

	return array;
}

/*!
 * \brief Allocate and fill a two-dimensional array of symbol pair counts.
 *
 * This function counts the occurrences of all possible symbol pairs in the
 * input sequence \p tis. The number of occurrences of \e xy is stored at row
 * \e x, column \e y of the returned array.
 *
 * \param buckets Number of buckets, i.e., the alphabet size.
 * \param tis Pointer to transformed input sequence.
 * \param suffixes Number of suffixes in the sequences, i.e., its length.
 * \param special_bucket Index of the special bucket.
 *
 * \returns Pointer to a two-dimensional \p buckets x \p buckets array.
 */
static size_t *count_symbol_pairs(fid_Uint16 buckets, const fid_Symbol *tis,
																	UINT suffixes, fid_Uint16 special_bucket)
{
	size_t *sub_bucket_sizes=malloc_safe(sizeof(size_t)*buckets*buckets);
	assert(sub_bucket_sizes != NULL);
	memset(sub_bucket_sizes,0,sizeof(size_t)*buckets*buckets);

	fid_Uint16 m=fid_REGULARSYMBOL(*tis)?(fid_Uint16)*tis:special_bucket;
	fid_Uint16 n;
	const fid_Symbol *max=tis+suffixes+2;

	/* Count the elements in each bucket and sub-bucket */
	for(++tis; tis < max; ++tis)
	{
		/* All special symbols share one special bucket. */
		n=fid_REGULARSYMBOL(*tis)?(fid_Uint16)*tis:special_bucket;
		sub_bucket_sizes[m*buckets+n]++;
		m=n;
	}

	return sub_bucket_sizes;
}

/*!
 * \brief Allocate and fill pointers to subbuckets.
 *
 * This function also allocates and initializes an array of bucket sizes.
 *
 * \param sub_bucket_sizes The sizes of all subbuckets as computed by
 * count_symbol_pairs().
 * \param buckets Number of buckets, i.e., the alphabet size.
 * \param info The bucket sizes and the size of the largest subbucket are
 *	stored within this structure.
 * \param suftab Pointer to the suffix array.
 *
 * \returns Pointer to two-dimensional \p buckets x \p buckets array (Iliffe
 * style), with each entry pointing into suftab. The bucket sizes are returned
 * in \p bucket_sizes.
 */
static UINT*** make_subbucket_pointers(const size_t *sub_bucket_sizes,
																			 fid_Uint16 buckets,
																			 bucket_info_t *const info, UINT *suftab)
{
	/* Array for storing the bucket sizes */
	size_t *bsizes = info->bucket_sizes = malloc_safe(sizeof(size_t) * buckets);
	assert(bsizes != NULL);
	memset(bsizes, 0, sizeof(size_t) * buckets);

	/* Stores the size of the largest known subbucket */
	size_t largest_sbucket_size = 0;

	/* Allocate memory for the pointers to the sub-buckets */
	UINT*** sbuckets = alloc_buckets(buckets); 
	assert(sbuckets != NULL);


	/* Set up pointers to the beginning of each sub-bucket in the mapped
	  suffix array file. */
	for(fid_Uint16 i = 0; i < buckets; i++) {
		for(fid_Uint16 j = 0; j < buckets; j++) {
			/* Initialize the iterator for each sub-bucket, this pointer
			 * points to the next un-used slot in each sub-bucket. */
			sbuckets[i][j] = suftab;
			
			if(largest_sbucket_size < *sub_bucket_sizes) {
				/* Update size information */
				largest_sbucket_size = *sub_bucket_sizes;
			}

			/* Calculate the size of each bucket from the size of its sub-buckets. */
			bsizes[i] += *sub_bucket_sizes;

			/* Move to the next sub-bucket. */
			suftab += *sub_bucket_sizes;

			++sub_bucket_sizes;
		}
	}

	info->largest_sbucket_size = largest_sbucket_size;

	return sbuckets;
}

/*!
 * \brief Fill subbuckets with correct data.
 *
 * This function actually presets the suffix array so that suffixes with an
 * lcp of 2 are grouped.
 *
 * \param sub_buckets Two-dimensional array of pointers into suftab.
 * \param buckets Number of buckets, i.e., the alphabet size.
 * \param tis Pointer to transformed input sequence.
 * \param suffixes Number of suffixes in the sequences, i.e., its length.
 * \param special_bucket Index of the special bucket.
 */
static void fill_subbuckets(UINT ** const *sub_buckets, fid_Uint16 buckets,
														const fid_Symbol *tis, UINT suffixes,
														fid_Uint16 special_bucket)
{
	/* Temporary pointers to the next free position in each sub-bucket,
	 * copy from sub_buckets */
	UINT ***sbuckets_next = alloc_buckets(buckets);
	assert(sbuckets_next != NULL);
	memcpy(sbuckets_next+buckets,sub_buckets+buckets,
				 sizeof(UINT **)*buckets*buckets);

	/* Re-set m to the first symbol of the first suffix */
	fid_Uint16 m=fid_REGULARSYMBOL(tis[0])?(fid_Uint16)tis[0]:special_bucket;
	fid_Uint16 n;

	/* Loop over the tis file once more, copy each suffix array index to
	  the next free slot in the correct bucket. */
	for(UINT i = 0; i <= suffixes; i++) {
		/* Second character of this suffix */
		n=fid_REGULARSYMBOL(tis[i+1])?(fid_Uint16)tis[i+1]:special_bucket;

		/* Insert into the next free slot of the correct bucket */
		*((sbuckets_next[m][n])++)=i;

		m = n;
	}

	free(sbuckets_next);
}

static const bucket_info_t *global_info_ptr_for_qsort;

static int compare_bucket_sizes(const void *a, const void *b)
{
	return global_info_ptr_for_qsort->bucket_sizes[*(const fid_Symbol *)a]-
				 global_info_ptr_for_qsort->bucket_sizes[*(const fid_Symbol *)b];
}

static void compute_sort_order(bucket_info_t *const info)
{
	for(fid_Uint16 i=0; i < info->bucket_count-1; ++i) info->sort_order[i]=i;

	global_info_ptr_for_qsort=info;
	qsort(info->sort_order,info->bucket_count-1,sizeof(fid_Symbol),
				compare_bucket_sizes);
	global_info_ptr_for_qsort=NULL;
}

/*!
 * \brief Initializes the suffix array and buckets.
 *
 * The transformed input sequence is read and its suffixes are
 * sorted into the initial buckets and sub-buckets using counting
 * sort. The suffixes are then written to the suffix array file in
 * sub-bucket order.
 * \param esa Initialized suffix array structure which contains
 * 	pointers to the required data in memory.
 * \param info Newly initialized bucket info structure, information
 * 	about the resulting buckets will be written to this structure.
 */
void buckets_tis_to_suf(fid_Suffixarray *const esa, bucket_info_t *const info) {
	assert(info != NULL);
	assert(esa != NULL);
	
	/* Number of buckets = number of symbols */
	info->bucket_count = esa->alpha.num_of_syms;
	
	/* Initialize the arrays used for marking sorted buckets. */
	size_t bits=info->bucket_count;
	info->bucket_sorted = malloc_safe(BITVEC_BYTES(bits));
	assert(info->bucket_sorted != NULL);
	memset(info->bucket_sorted, 0, BITVEC_BYTES(bits));

	bits*=info->bucket_count;
	info->sbucket_sorted = malloc_safe(BITVEC_BYTES(bits));
	assert(info->sbucket_sorted != NULL);
	memset(info->sbucket_sorted, 0, BITVEC_BYTES(bits));

	/*
	 * Regular characters are considered to be lexographically
	 * smaller than special characters. Therefore all special
	 * characters are sorted into the last bucket/sub-bucket.
	 */
	fid_Uint16 special_bucket = info->bucket_count - 1;

	/* 2D Array used for determining the size of each sub-bucket */
	size_t *sub_bucket_sizes=
		count_symbol_pairs(info->bucket_count,esa->sequences.tisfile.content,
											 esa->sequences.total_length.VU,special_bucket);

	info->sub_buckets=
		make_subbucket_pointers(sub_bucket_sizes, info->bucket_count,
														info, (UINT *)esa->suffile.content);

	free(sub_bucket_sizes);

	fill_subbuckets(info->sub_buckets,info->bucket_count,
									esa->sequences.tisfile.content,
									esa->sequences.total_length.VU,special_bucket);

	compute_sort_order(info);
}

/*!
 * \brief Initializes a bucket_info_t structure.
 * 
 * \param info Pointer to an un-initialized info structure.
 */
void bucket_info_init(bucket_info_t *const info) {
	if(info == NULL) {
		return;
	}

	info->bucket_count = 0;

	info->bucket_sizes = NULL;

	info->sub_buckets = NULL;

	info->bucket_sorted = NULL;

	info->sbucket_sorted = NULL;

	info->largest_sbucket_size = 0;
}

/*! 
 * \brief Frees the memory used by a bucket_info_t structure.
 *
 * This function frees all the memory used by a bucket_ifno
 * structure, leaving the structure in an undefined state.
 * \param info Pointer to a bucket_info_t structure.
 */
void bucket_info_free(bucket_info_t* info) {
	if(info == NULL) {
		return;
	}

	if(info->bucket_sizes != NULL) {
		free(info->bucket_sizes);
	}
	
	if(info->bucket_sorted != NULL) {
		free(info->bucket_sorted);
	}

	if(info->sbucket_sorted != NULL) {
		free(info->sbucket_sorted);
	}

	if(info->sub_buckets != NULL) {
		free(info->sub_buckets);
	}
}

static void forward_copy(const bucket_info_t *const info,
			 UINT *source_p, fid_Symbol sym,
			 UINT *target_buckets[], const fid_Symbol *seq)
{
	assert(source_p <= target_buckets[sym]);

	fid_Symbol prev_symbol;

	for(/* nothing */; source_p < target_buckets[sym]; ++source_p) {

		/* Skip the first suffix since it has no previous symbol */
		/* TODO: Checking this for every suffix is rather inefficient.
		 * It'd be smarter to check if this bucket actually contains the
		 * first suffix beforehand. */
		if(*source_p == 0) {
			continue;
		}

		prev_symbol = seq[*source_p - 1];

		/* Don't sort the special bucket using this method
			 since the special bucket is handled separately. */
		if(fid_SPECIALSYMBOL(prev_symbol)) {
			continue;
		}

		if(IS_WHOLE_BUCKET_SORTED(info,prev_symbol))
		{
			continue;
		}

		/* Select the target bucket by looking at the previous symbol
		  and store the suffix into the next free slot in the target
		  bucket. */
		*(target_buckets[prev_symbol]) = *source_p - 1;
		
		target_buckets[prev_symbol]++; 
	}
}

static void backward_copy(const bucket_info_t *const info,
			  UINT *source_p, fid_Symbol sym,
			  UINT *target_buckets[], const fid_Symbol *seq)
{
	assert(source_p >= target_buckets[sym]);

	fid_Symbol prev_symbol;

	for(/* nothing */; source_p > target_buckets[sym]; --source_p) {

		/* Skip the first suffix since it has no previous symbol */
		/* TODO: Checking this for every suffix is rather inefficient.
		 * It'd be smarter to check if this bucket actually contains the
		 * first suffix beforehand. */
		if(*source_p == 0) {
			continue;
		}

		prev_symbol = seq[*source_p - 1];

		/* Don't sort the special bucket using this method
			 since the special bucket is handled separately. */
		if(fid_SPECIALSYMBOL(prev_symbol)) {
			continue;
		}

		if(IS_WHOLE_BUCKET_SORTED(info,prev_symbol))
		{
			continue;
		}

		/* Select the target bucket by looking at the previous symbol
		  and store the suffix into the next free slot in the target
		  bucket. */
		*(target_buckets[prev_symbol]) = *source_p - 1;
		
		target_buckets[prev_symbol]--; 
	}
}

/*
 * \brief Uses the newly sorted bucket B(a) to quickly sort the 
 * 	sub-buckets B(*, a) using copy sort.
 *
 * If the bucket B(a) is known to be sorted we can walk over
 * the B(a) bucket. For each suffix of the type "ba...."
 * we can now insert this suffix into the next free slot of the
 * B(b, a) sub-bucket. At the end of this procedure all
 * B(*, a) sub-buckets will be sorted.
 * \param esa Suffix array structure.
 * \param info bucket_info_t structure containing bucket information.
 * \param source The number of the newly-sorted bucket that will
 * 	be used as the source for copy sort.
 */
void copy_sort(fid_Suffixarray *const esa, bucket_info_t *const info,
		const fid_Uint16 source) {
	assert(esa != NULL);
	assert(info != NULL);
	assert(source < info->bucket_count);
	assert(info->sub_buckets != NULL);

	fid_Uint16 i;
	UINT* target_buckets[info->bucket_count];

	assert(!IS_WHOLE_BUCKET_SORTED(info,source));

	/* If we're copying the sorted Bucket B(a) then target_buckets
	  will contain pointers to the start of the sub-buckets b(a,a),
	  b(b,a), b(c,a) etc. These are the subbuckets that will be sorted
	  by looking at the source bucket. */

	/* Loop over all suffixes in the sorted bucket once. */
	if(info->sub_buckets[source][source] > info->sub_buckets[source][0])
	{
		for(i = 0; i < info->bucket_count; i++)
			target_buckets[i] = info->sub_buckets[i][source];
		forward_copy(info,info->sub_buckets[source][0],source,
                             target_buckets,esa->sequences.tisfile.content);
	}
	if(info->sub_buckets[source][source] < info->sub_buckets[source+1][0])
	{
		for(i = 0; i < info->bucket_count; i++)
			target_buckets[i] = info->sub_buckets[i][source+1]-1;
		backward_copy(info,info->sub_buckets[source+1][0]-1,source,
                              target_buckets,esa->sequences.tisfile.content);
	}

	for(i = 0; i < info->bucket_count; i++)
		SET_SBUCKET_SORTED(info,i,source);
}
