/*
	Copyright (C) 2008 David Fleer.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef _MKESA_ALGO_H_
#define _MKESA_ALGO_H_

#include <libfid.h>

#ifdef MKESA_DATA64
#include <libfid.64>
#else
#include <libfid.32>
#endif


/* Various returns of the comparison functions */
/* Returned if the order between two suffixes was decided
 * by comparing two special characters. */
#define CMP_SPECIAL 2
/* returned if the order between two suffixes was decided
 * by one suffix being shorter than the other */
#define CMP_LENGTH 3


void buckets_tis_to_suf(fid_Suffixarray *const esa, bucket_info_t *const info);

void bucket_info_init(bucket_info_t *const info);

void bucket_info_free(bucket_info_t *const info);

void copy_sort(fid_Suffixarray *const esa, bucket_info_t *const info, 
		const fid_Uint16 source);

#define RETURN_CMP_EQUAL(PTR1, PTR2) {\
	if(fid_REGULARSYMBOL(*(PTR1)) && fid_REGULARSYMBOL(*(PTR2)))\
	{\
		/* standard symbol comparison */\
		if(*(PTR1) < *(PTR2)) return -1;\
		if(*(PTR1) > *(PTR2)) return 1;\
		return 0;\
	}\
	\
	if(fid_REGULARSYMBOL(*(PTR1))) return -CMP_LENGTH;\
	if(fid_REGULARSYMBOL(*(PTR2))) return CMP_LENGTH;\
	return ((PTR1) < (PTR2)) ? -CMP_SPECIAL : CMP_SPECIAL;\
}

#define RETURN_CMP_DIFFERENT(PTR1, PTR2) {\
	if(fid_REGULARSYMBOL(*(PTR1)) && fid_REGULARSYMBOL(*(PTR2)))\
	{\
		/* standard symbol comparison */\
		return (*(PTR1) < *(PTR2)) ? -1 : 1;\
	}\
	\
	/* the suffix that continues with a regular symbol is smaller; if
	 * both characters are special symbols use the position of the
	 * suffixes to decide the order */\
	if(fid_REGULARSYMBOL(*(PTR1))) return -CMP_LENGTH;\
	if(fid_REGULARSYMBOL(*(PTR2))) return CMP_LENGTH;\
	return ((PTR1) < (PTR2)) ? -CMP_SPECIAL : CMP_SPECIAL;\
}

/*!
 * \brief Standard comparison between two suffixes using LCP.
 *
 * \param suffix1 Pointer to the beginning of the first suffix.
 * \param suffix2 Pointer to the beginning of the second suffix.
 * \param lcp Pointer to the length of the common prefix of both
 * 	suffixes. The prefix is skipped when comparing the suffixes.
 * 	After the function returns the lcp value will be updated
 * 	to represent the length of the longest common prefix of the
 * 	two suffixes.
 * \returns Smaller than zero if the first suffix was smaller than
 * 	the second, greater than zero otherwise.
 */
static inline int cmp_suf_lcp(const fid_Symbol *restrict suffix1,
		const fid_Symbol *restrict suffix2, UINT *const lcp) {
	const fid_Symbol *restrict old_suf1ptr=suffix1;

	while(*suffix1 == *suffix2 && fid_REGULARSYMBOL(*suffix1)) {
		++suffix1;
		++suffix2;
	}

	*lcp+=suffix1-old_suf1ptr;

	RETURN_CMP_DIFFERENT(suffix1, suffix2);
}

/*!
 * \brief Standard comparison between two suffixes using LCP.
 *
 * \param suffix1 Pointer to the beginning of the first suffix.
 * \param suffix2 Pointer to the beginning of the second suffix.
 * 	sequence.
 * \returns Smaller than zero if the first suffix was smaller than
 * 	the second, greater than zero otherwise.
 */
static inline int cmp_suf_lcp_ro(const fid_Symbol *restrict suffix1,
		const fid_Symbol *restrict suffix2) {
	while(*suffix1 == *suffix2 && fid_REGULARSYMBOL(*suffix1))  {
		++suffix1;
		++suffix2;
	} 
	RETURN_CMP_DIFFERENT(suffix1, suffix2);
}


/*!
 * \brief Comparison of a certain character of two suffixes.
 *
 * \param suffix1 Pointer to the beginning of the first suffix.
 * \param suffix2 Pointer to the beginning of the second suffix.
 * \returns Smaller than zero if the character of first suffix was 
 * 	smaller than the second, greater than zero otherwise,
 * 	zero if both characters were equal.
 */
static inline int cmp_suf_depth(
		const fid_Symbol *restrict suffix1,
		const fid_Symbol *restrict suffix2) {

		RETURN_CMP_EQUAL(suffix1, suffix2);
}
#endif

