/*
	Copyright (C) 2008, 2009 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef MULTITHREADED
#include <pthread.h>
#endif
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif

#include "common.h"
#include "blind.h"
#include "algo.h"
#include "util.h"

/*!
 * \brief Initializes a new blind trie node.
 *
 * All relevant pointers of the node will be set to NULL.
 * \param node The node that will be initialized.
 */
static inline void init_bnode(blind_node_t *const node) {
	assert(node != NULL);

	node->children = NULL;
	node->parent = NULL;
	node->next = NULL;
	node->leaves = NULL;
}

/*!
 * \brief Calculates the total length of a suffix.
 *
 * The total length of a suffix is the distance between its
 * first character at offset \p pos and the end of the input 
 * sequences. Special characters / seperators are not considered.
 * \param pos Position of the first character of the suffix.
 * \param tp Pointer to the beginning of the input sequence.
 * \param tm Pointer to the end of the input sequence.
 * \returns The length of the suffix starting at pos.
 */
static inline UINT suf_length(const UINT pos,
		const fid_Symbol* tp, const fid_Symbol* tm) {
	return (tm - tp) - pos;
}

/*!
 * \brief Sorts a set of suffixes using blind sort.
 *
 * First, an empty blind trie is created. Then all suffixes
 * are inserted one by one into the blind trie. The finished
 * blind trie is then traversed and the sorted suffixes
 * are written back into the suffix array.
 * \param a A pointer to the first suffix in the set.
 * \param n Number of suffixes in the set.
 * \param sp Sort parameters.
 * \param lcp The longest common prefix of all the suffixes in
 * 	the set. The first lcp symbols are ignored while blind sorting.
 */
void blind_sort(UINT* a, const size_t n, const sort_params_t* sp, 
		const UINT lcp) {
	if(n < 2) {
		/* Do not sort single-element tries */
		return;
	}
	
	/* Pre-allocate space for all leaves since we already know the exact count */
	blind_leaf_t* leaves = malloc_safe(sizeof(blind_leaf_t) * n);
	assert(leaves != NULL);

	/* Insert the nodes into a new trie */
	blind_node_t* root = build_trie(a, n, leaves, sp->tp, sp->tm, lcp);

	/* dump_trie(root, sp->tp, sp->tm); */
	
	/* Traverse the trie, copy the leaves into the suffix array */
	traverse_trie(root, &a);

	/* Destroy the trie & free its memory */
	destroy_trie(root);

	/* Free the leaves */
	free(leaves);
}

/*! 
 * \brief Frees all memory used by the nodes of a blind trie.
 *
 * First destroy_trie is called for all children of the 
 * root node, then the root node itself is free()'d.
 * Leaves, however, are not free()'d and have to be
 * cleaned up seperately.
 * \param root The root node of the trie that should be removed.
 */
void destroy_trie(blind_node_t* root) {
	assert(root != NULL);

	blind_node_t* child_p;

	while(root->children != NULL) {
		child_p = ((blind_node_t*) root->children)->next;
		
		destroy_trie(root->children);
		
		root->children = child_p;
	}

	free(root);
}

/*!
 * \brief Traverses a blind trie and copies its leaves to a
 * 	suffix array.
 *
 * First all the children of the root node are traversed recursively,
 * then the leaves belonging to the root node are copied to the
 * suffix array.
 * \param root The root node of the trie.
 * \param sufarray The beginning of the suffix array.
 */
void traverse_trie(const blind_node_t* root, UINT** sufarray) {
	assert(root != NULL);
	assert(sufarray != NULL);
	assert(*sufarray != NULL);

	/* Get pointers to the leaves & children */
	blind_leaf_t* leaves_p = root->leaves;
	blind_node_t* child_p = root->children;

	/* Traverse all the child tries */
	while(child_p != NULL) {
		traverse_trie(child_p, sufarray);
		child_p = child_p->next;
	}

	/* Copy all the leaves to the resulting suffix array */
	while(leaves_p != NULL) {
		(**sufarray) = leaves_p->suffix;
		(*sufarray)++;
		leaves_p = leaves_p->next;
	}
}	

void dump_trie(const blind_node_t* root, const fid_Symbol* tp,
		const fid_Symbol* tm) {
	assert(root != NULL);
	blind_node_t* child = root->children;
	blind_leaf_t* leaf = root->leaves;
	const fid_Symbol* text;
	UINT n = 0;
	
	printf("Node: Symbol: "fid_SYMFMT" LCP: "UINTFMT"\n",
		root->symbol, root->lcp);

	while(leaf != NULL) {
		printf("Leaf: 0x%lx \"", (unsigned long) leaf->suffix);
		text = tp + leaf->suffix;
		while(text < tm) {
			printf("%x ", *(text++));
		}
		printf("\"\n");
		leaf = leaf->next;
	}
	while(child != NULL) {
		printf("\nEntering child "UINTFMT"\n", n);
		dump_trie(child, tp, tm);
		printf("\nReturning from child "UINTFMT"\n", n++);
		child = child->next;
	}
}


/*!
 * \brief Constructs a blind trie from a set of suffixes.
 *
 * Takes an array of suffixes of length n and inserts them into
 *	a blind trie in the right order.
 *	See Feragina/Grossi for details on the steps involved.
 * \param sufarray Pointer to the first suffix in the set of
 * 	suffixes.
 * \param n Number of suffixes in the set.
 * \param leaves Pointer to an array of n blind_leaf_t.
 * \param tp Pointer to the beginning of the input sequence.
 * \param tm Pointer to the end of the input sequence.
 * \param lcp The longest common prefix of all the suffixes
 * 	in the set. The first lcp symbols are ignored when constructing
 * 	the trie.
 */
blind_node_t* build_trie(UINT* sufarray, const size_t n,
		blind_leaf_t *const leaves, const fid_Symbol* tp, 
		const fid_Symbol* tm, const UINT lcp) {

	assert(sufarray != NULL);
	assert(tp != NULL);
	assert(tm != NULL);

	size_t i;
	blind_node_t *root, *node_p, *target, *branch;
	blind_leaf_t* leaf;
	const fid_Symbol *suffix, *suffix_p;
	int cmp;
	UINT suf_lcp;

	const fid_Symbol* base = tp + lcp;

	
	/* Manually insert the first element to avoid FUBAR */
	
	/* Create the root node */
	root = malloc_safe(sizeof(blind_node_t));
	assert(root != NULL);

	init_bnode(root);
	root->lcp = 0;
	root->symbol = 0;

	/* Create a node for the first element */
	node_p = malloc_safe(sizeof(blind_node_t));
	assert(root != NULL);

	init_bnode(node_p);

	/* Initialize the first node */
	node_p->symbol = *(tp + sufarray[0] + lcp);
	node_p->lcp = suf_length(sufarray[0] + lcp, tp, tm);
	node_p->leaves = leaves;
	leaves[0].suffix = sufarray[0];
	leaves[0].next = NULL;

	/* Attach first node to the root node */
	node_p->parent = root;
	root->children = node_p;
	
	

	/* Insert each suffix into the trie */
	for(i = 1; i < n; i++) {
		/* dump_trie(root, tp, tm); */
		suffix = tp + sufarray[i] + lcp;
		suffix_p = suffix;

		target = root;	

		/* First phase of the blind search, try to trace a
		  path to the target node. */
		while(target->children != NULL) {

			/* Try to find the child of the current node that has a matching
			  edge symbol. */
			node_p = target->children;

			/* Skip all the smaller symbols. */
			if(fid_REGULARSYMBOL(*suffix_p))
			{
				while(node_p->symbol < *suffix_p && node_p->next != NULL) {
					node_p = node_p->next;
				}

				if(node_p->symbol != *suffix_p) {
					/* Since the children are sorted according to their symbols,
						 reaching this points means that there's no matching child. */
					break;
				} else if((suffix + node_p->lcp) >= tm) {
					/* Abort if the next node would lead to out-of-bounds
						 comparison. (This is as deep as we're going to get) */
					target = node_p;
					break;
				} else {
					/* Use next node */
					target = node_p;
					suffix_p = suffix + target->lcp;
				}
			}
			else
			{
				/* just like above, but with a twist */
				while(fid_REGULARSYMBOL(node_p->symbol) && node_p->next != NULL) {
					node_p = node_p->next;
				}

				if(fid_REGULARSYMBOL(node_p->symbol)) {
					break;
				} else if((suffix + node_p->lcp) >= tm) {
					target = node_p;
					break;
				} else {
					target = node_p;
					suffix_p = suffix + target->lcp;
				}
			}
		}

		/* At this point we're at the deepest node for which the symbols
		  associated with the edges in the path from the root to the node
		  match the suffix.
		  We now have to find a leaf that belongs to the target node or to
		  one of its children. */
		
		while(target->leaves == NULL) {
			/* All terminal nodes must have leaves, therefore a node without
			  leaves must have at least a single child. */

			/* Go to the first child. */
			target = target->children;
		}
	
		/* We found a leaf */
		leaf = target->leaves;

		/* Step one complete, we now have the required leaf.
		  Now follows step two, were we compare the leaf suffix with the
		  given one to find out where the given suffix should be inserted
		  into the trie. */

		/* Initialize a new leaf with the suffix that we want to insert */
		leaves[i].suffix = sufarray[i];
		leaves[i].next = NULL;

		suf_lcp = lcp;
		cmp = cmp_suf_lcp(base + sufarray[i],
				base + leaf->suffix, &suf_lcp);
		/* We only care about the relative lcp, not the total lcp */
		suf_lcp -= lcp;


		switch(cmp) {
			case -CMP_SPECIAL:
				/* The compared suffixes are equal up to a special character,
				  so we insert the suffix as a leaf of the current node. */
				leaves[i].next = target->leaves;
				target->leaves = &(leaves[i]);

				break;

			case CMP_SPECIAL:
				/* Insert the suffix as a leaf of the current node 
				  while maintaining leaf order. */
				while(leaf->next != NULL) {
					if(((blind_leaf_t*) leaf->next)->suffix > sufarray[i]) {
						break;
					}
					leaf = leaf->next;
				}

				/* Insert suffix into the right position */
				leaves[i].next = leaf->next;
				leaf->next = &(leaves[i]);

				break;

			default:
				/* Find hit node, create new node, insert suffix as leaf of node. */

				/* Find the hit-node */
				while(target->parent != NULL) {
					/* Move upwards until we find the last node before the "error"
					  in the comparison of the new string and the old leaf. */
					if(((blind_node_t*) target->parent)->lcp < suf_lcp) {
						break;
					}
					target = target->parent;
				}

				if(target->lcp == suf_lcp) {
					/* If the hit node LCP is equal to the LCP of the compared suffixes
					  then the new node/leaf will be a direct sub-node of the hit node. */
					if(cmp == CMP_LENGTH) {
						/* The new suffix was shorter, with its length being
						 * equal to the lcp of this node we can insert the new suffix
						 * as a direct leaf of the target node */
						leaf = target->leaves;

						/* Insert the suffix as a leaf of the current node
						  while maintaining leaf order. */
						if(leaf == NULL) {
							target->leaves = &(leaves[i]);
						} else {
							while(leaf->next != NULL) {
								if(((blind_leaf_t*) leaf->next)->suffix > sufarray[i]) {
									break;
								}
								leaf = leaf->next;
							}

							/* Insert suffix into the right position */
							leaves[i].next = leaf->next;
							leaf->next = &(leaves[i]);
						}
					} else {

						/* Create a new sub-node */
						node_p = malloc_safe(sizeof(blind_node_t));
						assert(node_p != NULL);

						init_bnode(node_p);
					
						node_p->parent = target;
						node_p->lcp = suf_length(sufarray[i] + lcp, tp, tm);
						node_p->symbol = *(tp + sufarray[i] + lcp + target->lcp);
						node_p->leaves = &(leaves[i]);
					
						if(target->children == NULL) {
							/* The new node is the first child of the hit node */
							target->children = node_p;

						} else {
							/* The hit node already has some children, insert
							  the new node while maintining order among the offspring */
							if(((blind_node_t*) target->children)->symbol > node_p->symbol) {
								node_p->next = target->children;
								target->children = node_p;
								
							} else {
								target = target->children;
								
								while(target->next != NULL) {
									if(((blind_node_t*) target->next)->symbol > node_p->symbol) {
										break;
									}
									target = target->next;
								}
						
								node_p->next = target->next;
								target->next = node_p;
							}
						}
					}
				} else {
					/* Creating two new nodes */

					/* Create the first node, the hit node and the other */
					/* new node will be children of this node */
					branch = malloc_safe(sizeof(blind_node_t));
					assert(branch != NULL);

					init_bnode(branch);
					
					branch->parent = target->parent;
					branch->lcp = suf_lcp;
					branch->symbol = target->symbol;

					/* Replace the old node as a child of the parent node. */
					if(((blind_node_t*) branch->parent)->children == target) {
						/* The old node was the first child of the parent */
						((blind_node_t*) branch->parent)->children = branch;
					} else {
						node_p = ((blind_node_t*) branch->parent)->children;
						while(node_p->next != target) {
							node_p = node_p->next;
						}
						node_p->next = branch;
					}
					branch->next = target->next;

					/* Change the hit node to reflect the new trie structure */
					target->symbol = *(tp + leaf->suffix + lcp + suf_lcp);
					target->parent = branch;

					if(cmp == CMP_LENGTH) {
						/* The new suffix is a leaf of the new node */
						branch->leaves = &(leaves[i]);
						branch->children = target;
						/* The target node is now the only child of the new node */
						target->next = NULL;
					} else {

						/* Create the second node, the inserted suffix will be */
						/* a leaf of this node */
						node_p = malloc_safe(sizeof(blind_node_t));
						assert(node_p != NULL);

						init_bnode(node_p);

						/* Initialize the second node */
						node_p->parent = branch;
						node_p->lcp = suf_length(sufarray[i] + lcp, tp, tm);
						node_p->symbol = *(tp + sufarray[i] + lcp + suf_lcp);
						node_p->leaves = &(leaves[i]);

						if(cmp < 0) {
							/* The new node is smaller than the hit node */
							branch->children = node_p;
							node_p->next = target;
							target->next = NULL;

						} else {
							/* The new node is larger than the hit node */
							branch->children = target;
							target->next = node_p;
						}
					
					}
					
				break;
			}
		}
	}

	return root;
}

