/*
	Copyright (C) 2008, 2009 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h> 
#ifdef HAVE_LIBGEN_H
#include <libgen.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "common.h"
#include "cli.h" 
#include "util.h" 

#ifdef MULTITHREADED
#include "threads.h"
#endif

/* cli.c: Command line interface handling for mkesa */

/*!
 * \brief Prints a version message.
 */
static void print_version(void) {
	printf("%s - Enhanced suffix array construction tool.\n"
"Copyright (C) 2008, 2009  David Fleer, Robert Homann.\n"
"Please send feedback or bug reports to %s.\n\n"
"This program is free software; you can redistribute it and/or modify it under\n"
"the terms of the GNU General Public License as published by the Free Software\n"
"Foundation; either version 2, or (at your option) any later version.\n\n"
"This program is distributed in the hope that it will be useful, but WITHOUT\n"
"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n"
"FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.\n"
"You should have received a copy of the GNU General Public License along with\n"
"this program (see the file COPYING); if not, write to the Free Software\n"
"Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n\n"
"Compiled for %d bit architecture (" TARGET_SYS ")"
#ifdef DEBUG
" (debug version)"
#endif
".\n",
PACKAGE_STRING,PACKAGE_BUGREPORT,fid_WORDSIZE);
	return;
}

/*!
 * \brief Prints the extended version message.
 *
 * Prints the standard version message followed
 * by information about the bit width of this version.
 * Only used when one of the different versiosn is
 * executed directly.
 */
static void print_ext_version(void) {
	print_version();

	printf("This program version operates on "
	#ifdef MKESA_DATA32
				 "32"
	#else
	#ifndef MKESA_DATA64
	#error Expected MKESA_DATA64.
	#endif
				 "64"
	#endif
				 " bit data structures.\n");
}

/*! 
 * \brief Initialize the given structure with default options.
 *
 * \param params The structure that should be initialized.
 */ 
void init_mkesa_params(mkesa_params_t* params) {
	params->verbose = 0;

	/* Pointers to the file names */
	params->al1_file = NULL;
	params->data_file = NULL;
	params->prj_name = NULL;

 	/* UINT sizes */
	params->uint_size = DEFAULT_UINT_SIZE;

	params->builtin_alphabet = fid_ALPHABET_DNA;
	params->use_builtin_alphabet = 0;

	params->update_mode = 1;
	
	/* Set default sorting parameters */
	params->maxdepth = DEFAULT_MAXDEPTH;
	params->segment_size = DEFAULT_SEGMENT_SIZE;

	/* Set default options for optional tables */
	params->build_sti = DEFAULT_BUILD_STI;
	params->build_lcp = DEFAULT_BUILD_LCP;
	params->build_skp = DEFAULT_BUILD_LCP;
	params->build_ois = 0;
	params->write_suf = 0;

	#ifdef MULTITHREADED
	params->threaded = DEFAULT_THREADED;
	#ifdef HAVE_SYSCONF
		params->thread_count = get_num_of_cpus();
		if(params->thread_count == 0) {
			/* Use default thread count if unable to determine
			 * the number of CPUs / cores */
			params->thread_count = DEFAULT_THREAD_COUNT;
		}
	#else
		params->thread_count = DEFAULT_THREAD_COUNT;
	#endif
	#endif
}

/*!
 * \brief Parses the list of optional tables passed using the -g switch.
 *
 * The table list is comma-seperated, the appropriate member of the params structure
 * is set for each entry of the list. An error message is printed if a malformed
 * list is encountered.
 * \param arg The table list as specified by the user.
 * \param params appropriate fields of this structure are set for each list entry.
 * \returns 0 on success, -1 on parse error or illegal table.
 */
static int parse_generate(const char* arg, mkesa_params_t* const  params) {
	assert(arg != NULL);
	assert(params != NULL);
	const char *p = arg;

	/* Advance to the next delimiter or end of string */
	while((*p != '\0') && (*p != ',')) {
		p++;
	}

	/* All table identifiers are exactly 3 characters long, so
	 * do some error handling if we encounter any with a different length */
	if((p - arg) != 3) {
		/* Ignore empty identifiers, error out otherwise */
		if((p - arg) != 0) {
			fprintf(stderr, "Error: Invalid table type for -g, \
try --help for valid output tables\n");
			return -1;
		}
	} else {

		/* Check the current list entry */
		if(strncmp(arg, "lcp", 3) == 0) {
			params->build_lcp = 1;
		} else if(strncmp(arg, "sti", 3) == 0) {
			params->build_sti = 1;
		} else if(strncmp(arg, "skp", 3) == 0) {
			params->build_lcp = 1;
			params->build_skp = 1;
		} else if(strncmp(arg, "ois", 3) == 0) {
			params->build_ois = 1;
		} else if(strncmp(arg, "suf", 3) != 0 && strncmp(arg, "tis", 3) != 0) {
			/* Unknown entry -> Error */
			fprintf(stderr, "Error: Invalid table type for -g, \
try --help for valid output tables\n");
			return -1;
		}

	}

	if(*p != '\0') {
		/* If we haven't reached the end of the list then repeat */
		return parse_generate(p + 1, params);
	}

	return 0;
}

/*!
 * \brief Parse command line options.
 *
 * This function parses the command line options as passed to main()
 * and stores the result as a mkesa_params_t struct.
 * Options in \p params that are not affected by the given 
 * command line options  are set to their defaults.
 *
 * \param argc, argv The command line options as passed to main().
 * \param params The structure that will be used for storing the result.
 * Will be initialized using init_mkesa_params.
 * \returns 0 on success, -1 if a critical error was encountered,
 * -2 if no error was encountered but the program should terminate 
 *  anyway (e.g. --help was specified).
 */ 
int parse_mkesa_params(int argc, char* argv[], mkesa_params_t* const params) {
	assert(params != NULL);
	assert(argv != NULL);
	assert(argc > 0);

	int r = 0;
	/* If this is != 0, then this mkesa has been called
		 by the launcher program. If it's 0 then
		 mkesa has been called direclty */
	int launched = 0;

	const char* optstring = 
	#ifdef MULTITHREADED
		"thvVc:a:p:d:b:s:e:g:lw:";
	#else
		"hvVa:p:d:b:s:e:g:lw:";
	#endif
	
	/* Initialize list of legal command line switches */
	static struct option longopts[] = {
	#ifdef MULTITHREADED
		{"threaded", no_argument, 0, 't'},
		{"threadcount", required_argument, 0, 'c'},
	#endif
		{"help", no_argument, 0, 'h'},
		{"segment-size", required_argument, 0, 'e'},
		{"verbose", no_argument, 0, 'v'},
		{"version", no_argument, 0, 'V'},
		{"al1", required_argument, 0, 'a'},
		{"project", required_argument, 0, 'p'},
		{"data", required_argument, 0, 'd'},
		{"generate", required_argument, 0, 'g'},
		{0, 0, 0, 0}
	};

	init_mkesa_params(params);

	fid_Uint32 temp;

	while(r != -1) {
		r = getopt_long(argc, argv, optstring, longopts, NULL);
		/* Process each command line switch */
		switch(r) {
			case -1:
				break;

			case 'h':
				/* Display the appropriate help message. */
				if(launched) {
					print_ext_help();
				} else {
					print_help();
				}
				return -2;

			case 'w':
				/* -w should only be used together with the launcher
					 program so we warn the user and abort. */
				if(!launched) {
					fprintf(stderr,
                  "Error: switch \"-w\" can only be used in conjunction with the mkesa launcher\n"
                  "program, but not when running this program directly. Please consider using the\n"
                  "mkesa program instead of mkesa32 or mkesa64 in order to specify the bit width.\n");
					return -2;
				}

			case 'V':
				/* Display the appropriate version message. */
				if(!launched) {
					print_ext_version();
				} else {
					print_version();
				}
				return -2;

			case '?':
			case ':':
				/* Unknown/illegal option, print error message.
				 * getopt() itself will also print a basic error message. */
				printf("Try --help for usage information\n");
				return -1;

			case 'v':	
				params->verbose++;
				break;
	#ifdef MULTITHREADED
			/* -t or -c are only allowed when using a version with 
			 * multithreading support. */
			case 't':
				params->threaded = 1;
				break;

			case 'c':
				if(fid_utils_parse_uint_32(optarg,&temp,NULL) == -1 ||
					 temp < 1 || temp > 255)
				{
					fprintf(stderr, "Error: \"-c\" expects an integer between 1 and 255 "
									"as its argument. (Given: %s)\n", optarg);
					printf("Try --help for usage information\n");
					return -1;
				}

				params->threaded = 1;
				params->thread_count = temp;
				break;
			
	#endif

			/* Switches for specifying various files. */
			case 'a':
				params->al1_file = optarg;
				break;

			case 'p':
				params->prj_name = optarg;
				break;

			case 'd':
				params->data_file = optarg;
				params->update_mode = 0;
				break;

			case 'b':
				switch(optarg[0]) {
					case 'd':
					case 'D':
						params->builtin_alphabet = fid_ALPHABET_DNA;
						break;

					case 'r':
					case 'R':
						params->builtin_alphabet = fid_ALPHABET_RNA;
						break;

					case 'n':
					case 'N':
						params->builtin_alphabet = fid_ALPHABET_DNARNA;
						break;

					case 'p':
					case 'P':
						params->builtin_alphabet = fid_ALPHABET_PROTEIN;
						break;

					default:
						fprintf(stderr, "Error: \"-b\" expects either \"d\" (DNA), \"r\" "
										"(RNA), \"n\" (mixed DNA and RNA), or\n"
										"\"p\" (protein) as built-in alphabet.\n"); 
						printf("Try --help for usage information\n");
						return -1;
				}
				params->use_builtin_alphabet = 1;
				break;

			case 's':
				if(fid_utils_parse_uint_32(optarg,&temp,NULL) == -1 || temp < 1)
				{
					fprintf(stderr, "Error: \"-s\" expects a positive non-zero integer "
									"as its argument. (Given: %s)\n", optarg);
					printf("Try --help for usage information\n");
					return -1;
				}

				params->maxdepth = temp;
				break;

			case 'e':
				if(fid_utils_parse_uint_32(optarg,&temp,NULL) == -1 ||
					 temp < 2 || temp > fid_UINT16_MAX)
				{
					fprintf(stderr, "Error: \"-e\" expects an integer between 2 and %hu "
									"as its argument. (Given: %s)\n", fid_UINT16_MAX, optarg);
					printf("Try --help for usage information\n");
					return -1;
				}

				params->segment_size = temp;
				break;

			case 'g':
				/* Call parse_generate to parse the list of additional tables */
				if(parse_generate(optarg, params) != 0) {
					return -1;
				}
				break;

			case 'l':
				/* Undocumented option that indicates
					 if we were called from the launcher program */
				launched = 1;
				break;

			default:	
				fprintf(stderr, 
					"Error: Internal error while parsing arguments\n");
				return -1;
		}
	}

	if(optind < argc)
	{
		int multiple=(argc-optind > 1);

		fprintf(stderr,"Error: Invalid option%s:",multiple?"s":"");
		if(multiple)
		{
			fputc('\n',stderr);
			while(optind < argc) fprintf(stderr,"  %s\n",argv[optind++]);
		}
		else fprintf(stderr," %s\n",argv[optind]);
		return -1;
	}

	if(!params->update_mode)
	{
		params->write_suf = 1;
		params->have_tis = 1;
	}
	else
	{
		if(params->prj_name != NULL && !test_file_readable(params->prj_name, "suf"))
			params->write_suf = 1;
		params->have_tis = (params->build_lcp || params->build_ois || params->write_suf);
	}

	return 0;
}

/* \brief Checks options for missing or conflicting settings.
 *
 * Error and/or warning messages are written to stderr/stdout.
 * The function will return with -1 when encountering a
 * critical error.
 * 
 * The following problems will generate an error:
 * - No input files specified (no --data).
 * The following problems will generate a warning:  
 * - No alphabet specified  (no --al1)
 * - No project name specified (no --project)
 *
 * \param params The mkesa_params_t struct that should be checked.
 * \returns 0 if no errors or warnings are encountered,
 * -1 if there was a critical error, -2 if one or more non-critical
 *  problems exist.  
 */  
int check_params_sanity(mkesa_params_t* params) {
	assert(params != NULL);

	int retcode = 0;

	if(params->data_file != NULL && params->al1_file == NULL 
			&& !params->use_builtin_alphabet) {
		/* If there's a raw data file specified  we need 
		 * either an alphabet file or internal alphabet so that we
		 * can parse/transform it. */
			fprintf(stderr, 
"Error: No valid alphabet file or builtin specified. Use -a to specify an\n\
alphabet file file or -b to use a built-in alphabet.\n\n");

			retcode = -1;
	} else if(params->prj_name == NULL) {
		/* Check for mandatory project name */
		fprintf(stderr,
"Error: No project name specified, please specify one\n\
using the -p or --project option.\n\n\
If you want to start from scratch with a new project file please\n\
specify a data file using the -d command.\n\n");
		
		retcode = -1;
	} else if(params->update_mode && 
			((params->al1_file != NULL) || (params->use_builtin_alphabet))) {
		/* No alphabets may be specified in update mode */
		fprintf(stderr,
"Error: Can't specify an alphabet (using -a or -b) in update mode.\n\
If you want to start from scratch using a specific alphabet, please\n\
specify a data file using the -d option.\n\n");
		retcode = -1;
	}

	if(retcode == -1 || (retcode == -2 && params->verbose)) {
		/* Display --help info message if any errors occurd or
		 * if warnings occurd and the verbose switch was enabled */
		printf("Try --help for usage information\n");	
	}

	return retcode;
}


void print_ext_help() {
	print_help();
	printf("-w, --width\tSpecifies the bit width for each entry of the suffix array.\n\
\t\tValid options are 32 (for 32 bit suffix arrays) or 64 (for 64\n\
\t\tbit suffix arrays). Be advised that chosing 32 bit mode on very\n\
\t\tlarge input files (roughly larger than 2^31) may result in\n\
\t\tundefined results since the suffix array entries might overrun.\n");
}

/*! \brief Print usage information.
 * 
 * This function writes basic usage information to stdout.
 * Information about threading-related switches is only shown
 * if threading was enabled at compile time.
 */ 
void print_help() {
	printf(
"Usage: mkesa -p project [-d DATAFILE (-a ALPHABET | -b BUILTIN)] [OPTION] ...\n\n");

	printf("%s - Enhanced suffix array construction tool.\n\
Copyright (C) 2008 David Fleer, Robert Homann.\n\n",
PACKAGE_STRING);

	printf("This program is distributed in the hope that it will be useful, but WITHOUT\n\
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n\
FOR A PARTICULAR PURPOSE. See mkesa -V for details and license information.\n\n");

	printf("Possible OPTIONs are:\n");

	printf("-p, --project\tSpecifies the project name.\n\
\t\tThis name will be used as the base name for all newly created\n\
\t\tfiles. If the program is used in \"update mode\", meaning without\n\
\t\tthe -d switch, then the program will look for existing files\n\
\t\tusing this name.\n\
\t\tExisting files will be used in \"update mode\" in order to build\n\
\t\tthe requested tables where possible. Therefore, a transformed\n\
\t\tinput sequence (TIS) and project file (PRJ) must exist before\n\
\t\tupdate mode can be used. If these files do not exist yet, use\n\
\t\tthe -d switch.\n");

	printf("-d, --data=DATAFILE\tSpecifies the data file.\n\
\t\tSequence data for which an enhanced suffix array is to be\n\
\t\tcomputed is read from this file. If this switch is used, then\n\
\t\tmkesa operates in \"create mode\". All requested files will be\n\
\t\tre-built from scratch, existing files will be overwritten.\n\
\t\tSupported file formats: FASTA, UniProt.\n");

	printf("-a, --al1=FILE\tSpecifies an alphabet file.\n\
\t\tThe alphabet is used for transforming the data file supplied\n\
\t\twith the -d switch.\n\
\t\tThis switch can only be used in \"create mode\" (i.e., in\n\
\t\tconjunction with the -d switch).\n");

	printf("-b BUILTIN\tSpecifies a built-in alphabet.\n\
\t\tPossible choices for BUILTIN are one of 'D'NA, 'R'NA,\n\
\t\t'N'ucleotide, and 'P'rotein.\n\
\t\tIf both -a and -b switches are given, then -a takes precedence.\n\
\t\tThis switch can only be used in \"create mode\" (i.e., in\n\
\t\tconjunction with the -d switch).\n");

	printf("-s\t\tSpecifies the maximum depth used for shallow sorting.\n\
\t\tAfter the specified depth is reached the sorting algorithm is\n\
\t\tswitched to depth sorting. (Default: %d.)\n", DEFAULT_MAXDEPTH);

	printf("-e, --segment-size\tSpecifies the segment size used for anchor sorting.\n\
\t\tGiven the anchor size d and input size n, anchor size requires\n\
\t\t6n/d additional bytes of memory. Must be between 2 and 65535.\n\
\t\t(Default: %d.)\n",
DEFAULT_SEGMENT_SIZE);

	printf("-g, --generate\tSpecifies what optional tables should be generated.\n\
\t\tValid table names are\n\
\t\t\t\"suf\" (suffix array, [*])\n\
\t\t\t\"sti\" (inverse suffix array)\n\
\t\t\t\"lcp\" (LCP information)\n\
\t\t\t\"skp\" (skip table, implies \"lcp\")\n\
\t\t\t\"tis\" (transformed input sequence, [*])\n\
\t\t\t\"ois\" (original input sequence, \"create mode\" only)\n\
\t\tMultiple table names can be specified as a comma-seperated\n\
\t\tlist, e.g., \"-g lcp,sti\". Tables marked with [*] are always\n\
\t\timplicitly specified and need not be given in the list. Also\n\
\t\tnote that existing tables are not re-built in \"update mode\".\n");

#ifdef MULTITHREADED
	/* Multithreading-specific help. */
	printf("-t, --threaded\tEnables threading.\n\
\t\tThe default number of worker threads used is %d.\n", DEFAULT_THREAD_COUNT);
	printf("-c, --threadcount\tEnables threading with the specified number of worker\n\
\t\tthreads (between 1 and 255).\n");
#endif

	printf("-v, --verbose\tEnables verbose mode.\n\
\t\tVerbosity can be increased by giving this option multiple\n\
\t\ttimes (like \"-vvv\").\n");
	printf("-h, --help\tDisplays this message.\n");
	printf("-V, --version\tDisplays version and copyright information.\n");

}
