/*
	Copyright (C) 2008 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef _MKESA_COMMON_H_
#define _MKESA_COMMON_H_

#include <libfid.h>

#ifdef MKESA_DATA64
#include <libfid.64>
#define DEFAULT_UINT_SIZE fid_UINTSIZE_64;
#else
#include <libfid.32>
#define DEFAULT_UINT_SIZE fid_UINTSIZE_32;
#endif

#define DEFAULT_MAXDEPTH 500
#define DEFAULT_SEGMENT_SIZE 500

#define DEFAULT_BUILD_LCP 0
#define DEFAULT_BUILD_STI	0

#ifdef MULTITHREADED 
#include <pthread.h>
#define DEFAULT_THREADED 0
#define DEFAULT_THREAD_COUNT 4
#endif

/* A UINT with only the MSB set, should be
 * platform (size of UINT / endianess)-independent. */
#define MSB_MASK (((UINT)-1 >> 1) + 1)

/* Compute number of bytes required to store I bits. */
#define BITVEC_BYTES(I)  (((I) >> 3) + 1)

/* Sets the I-th bit of a bit vector (represented by
 * an array of unsigned chars). */ 
#define BITVEC_SET(V,I)\
	((V)[(I) >> 3] |= (((unsigned char)1) << ((I)&0x07)))

/* Checks the I-th bit of a bit vector (represented by
 * an array of unsigned chars). */
#define BITVEC_TEST(V,I)\
	(((V)[(I) >> 3] & (((unsigned char)1) << ((I)&0x07))) != 0)

/*! Data structure storing various program settings,
	  many of them user-specified. */
typedef struct {
	/*! Indicates if verbose mode (extra output to stdout is used. */
	int verbose;

	#ifdef MULTITHREADED
		/*! Indicates if the program is used in multi-threaded mode. */
		int threaded;
		/*! Number of worker threads used in multi-threading mode. */
		fid_Uint16 thread_count;
	#endif

	/*! File name of the input alphabet file (if any). */
	char* al1_file;
	/*! File name of the input data file (if any). */
	char* data_file;
	/*! The name of the project as specified by the user. */
	char* prj_name;

	/*! Specifies the type of libfid built-in alphabet
		that should be used (if any). */
	fid_Alphabettype builtin_alphabet;
	/*! Specifies if a libfid built-in alphabet should be used. */
	int use_builtin_alphabet;

	/*! Indicates if the program is used in update mode. */
	int update_mode;

	/*! The size of a UINT (32 or 64 bit). */
	fid_Uintsize uint_size;
	/*! The size of the input file. */
	size_t input_file_size;

	/*! The maximum shallow sorting depth. */
	UINT maxdepth;
	/*! the size of each segment used for anchor sorting. */
	fid_Uint16 segment_size;

	/*! Indicates if the sti table 
	(inverse suffix table) should be built. */
	int build_sti;

	/*! Indicates if the lcp table
	(longest common prefix length) should be built. */
	int build_lcp;

	/*! Indicates if the skp table (skip table) should be built. */
	int build_skp;

	/*! Indicates if the ois table (original input sequence) should be built. */
	int build_ois;

	/*! Indicates if the suffix table needs to be re-written. */
	int write_suf;

	/*! True if the transformed input sequence is mapped. */
	int have_tis;
} mkesa_params_t;

/*! Data structure storing information about the buckets used
by bucket sort. */
typedef struct {
	/*! The number of buckets. */
	fid_Uint16 bucket_count;

	/*! Array containing the size of each bucket. */
	size_t* bucket_sizes;

	/*! Array that lists buckets in order of increasing size. */
	fid_Symbol sort_order[256];

	/*! Two-dimensional array containing a pointer to each
	sub-bucket. */
	UINT*** sub_buckets;

	/*! Bit vectors used for marking (sub-)buckets as sorted. */
	unsigned char* bucket_sorted;
	unsigned char* sbucket_sorted;

	/*! Size of the largest sub-bucket */
	size_t largest_sbucket_size;

} bucket_info_t;

#ifdef DEBUG
#define SET_WHOLE_BUCKET_SORTED(INFO,I) do\
{\
	assert(!BITVEC_TEST((INFO)->bucket_sorted,(I)));\
	BITVEC_SET((INFO)->bucket_sorted,(I));\
	for(fid_Uint16 priv_I=0; priv_I < (INFO)->bucket_count; ++priv_I)\
	{\
		if(!IS_SBUCKET_SORTED(INFO,(I),priv_I))\
		{\
			printf("INVALID: %d %d should have been sorted, but is not\n",\
						 (I),priv_I);\
			abort();\
		}\
	}\
} while(0)
#else /* !DEBUG */
#define SET_WHOLE_BUCKET_SORTED(INFO,I)  BITVEC_SET((INFO)->bucket_sorted,(I))
#endif /* DEBUG */

#define IS_WHOLE_BUCKET_SORTED(INFO,I)   BITVEC_TEST((INFO)->bucket_sorted,(I))

#define SET_SBUCKET_SORTED(INFO,I,J) do\
{\
	size_t bit=(I)*(INFO)->bucket_count+(J);\
	BITVEC_SET((INFO)->sbucket_sorted,bit);\
} while(0)

#define IS_SBUCKET_SORTED(INFO,I,J)\
	BITVEC_TEST((INFO)->sbucket_sorted,(I)*(INFO)->bucket_count+(J))

/*! Data structure containing various information used
by the sorting algorithms. */
typedef struct {
	/*! Pointer to the beginning of the transformed sequence data. */
	fid_Symbol *tp;

	/*! Pointer to the end of the transformed sequence data. */
	fid_Symbol *tm;

	/*! Pointer to the bucket information as generated by
	bucket sort.*/
	bucket_info_t* info;

	/*! The maximum shallow sorting depth. If this sorting depth
	is exceeded we'll switch to depth sort. */
	UINT maxdepth;

	/*! If the size of a group of unsorted suffixes is lower
	than this threshold then we'll sort them using blind sort. */
	UINT blind_threshold;

	/*! Number of segments used by anchor sorting. */
	UINT seg_count;

	/*! Pointer to each anchor within the
	suffix array for anchor sorting. */
	UINT **anchor;

	fid_Uint16 *offset;

	fid_Uint16 d;

	/*! Size of the largest sub-bucket */
	size_t largest_sbucket_size;

	/*! Bit vector for marking suffixes during
	induced sorting */
	unsigned char* marked_suffixes;

	/*! Size of the marked_suffixes bit vector */
	size_t marked_suffixes_size;

	#ifdef MULTITHREADED
	/*! Mutex used for protecting the anchor information
	in multithreaded mode. */
	pthread_mutex_t* anchor_mutex;
	/*! Indicates if thread-safe sorting should be used, e.g.
	if the anchor_mutex needs to be locked. */
	int threaded_sort;
	#endif
	
} sort_params_t;



#endif
