/*
	Copyright (C) 2008, 2009 David Fleer.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <stdio.h>

#include "convesa.h"

/* Macro for swapping two chars */
#define SWAP_CHAR(U, V) { char tmp; tmp = (U); (U) = (V); (V) = tmp; }

static void init_params(convesa_params_t* const params) {
	/* Initialize the parameters with sane default values */
	params->verbose = 0;

	params->input_file = NULL;
	params->output_file = NULL;

	params->in_endian = MKESA_UNDEF_ENDIAN;
	params->out_endian = MKESA_UNDEF_ENDIAN;

	params->in_uint_size = fid_UINTSIZE_UNDEF;
	params->out_uint_size = fid_UINTSIZE_UNDEF;
}

static void print_help(void) {
	/* Print usage information */
	printf("Usage: convesa -i INPUT_FILE -o OUTPUT_FILE -s INPUT_SIZE -e INPUT_ENDIANESS [-S OUTPUT_SIZE] [-E OUTPUT_ENDIANESS]\n\n");

	printf("convesa - Enhanced suffix array converter (part of %s).\nCopyright (C) 2008 David Fleer (Feedback to: %s).\n\n",
	PACKAGE_STRING,PACKAGE_BUGREPORT);

  printf("This program is distributed in the hope that it will be useful, but WITHOUT\n\
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n\
FOR A PARTICULAR PURPOSE.\n\n");

printf("-i, --input\tSpecifies the input file, has to be readable (mandatory).\n\
-o, --output\tSpecifies the output file, has to be writeable (mandatory).\n\
-e, --input-endian\tSpecifies the endianess of the input file.\n\
\t\tCan be either \"little\" or \"big\" (mandatory argument).\n\
-s, --input-size\tSpecifies the input file integer size.\n\
\t\tCan be either \"32\" or \"64\" (mandatory argument).\n\
-E, --output-endian\tSpecifies the endianess of the output file.\n\
\t\tIf this differs from the input endianess, convesa will switch\n\
\t\tthe endianess of the input.\n\
-S, --output-size\tSpecifies the output integer size. If this differs from\n\
\t\tthe input size, convesa will pad or cut the input accordingly.\n\
\t\tWarning: Information may be lost during 64->32 bit conversions.\n\
-v, --verbose\tEnables verbose mode.\n\
-h, --help\tDisplays this help message.\n");

}

static int parse_params(int argc, char* argv[], convesa_params_t* const params) {
	assert(params != NULL);
	assert(argv != NULL);
	assert(argc > 0);

	/* Initialize parameter struct */
	init_params(params);

	int r = 0;
	
	/* Valid options for getopt */
	const char* optstring = "i:o:e:E:s:S:vh";

	static struct option longopts[] = {
		{"input", required_argument, 0, 'i'},
		{"output", required_argument, 0, 'o'},
		{"input-endian", required_argument, 0, 'e'},
		{"output-endian", required_argument, 0, 'E'},
		{"input-size", required_argument, 0, 's'},
		{"output-size", required_argument, 0, 'S'},
		{"verbose", no_argument, 0, 'v'},
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};

	while(r != -1) {
		/* Parse the next switch */
		r = getopt_long(argc, argv, optstring, longopts, NULL);

		/* Handle the different options */
		switch(r) {
			case -1:
				break;

			case 'h':
				print_help();
				exit(EXIT_SUCCESS);
				break;


			case 'v':
				params->verbose++;
				break;
			case 'i':
				params->input_file = optarg;
				break;

			case 'o':
				params->output_file = optarg;
				break;

			case 'e':
				/* Input endianess switch */
				if(strncmp(optarg, "little", 6) == 0) {
					/* Little endian mode */
					params->in_endian = MKESA_LITTLE_ENDIAN;
				} else if(strncmp(optarg, "big", 3) == 0) {
					/* Big endian mode */
					params->in_endian = MKESA_BIG_ENDIAN;
				} else {
					fprintf(stderr, "Error: Invalid endianess: %s.\
(Valid: 'little' or 'big')\nTry --help for details.\n", optarg);
					exit(EXIT_FAILURE);
				}
			break;

			case 'E':
				/* Output endianess switch */
				if(strncmp(optarg, "little", 6) == 0) {
					params->out_endian = MKESA_LITTLE_ENDIAN;

				} else if(strncmp(optarg, "big", 3) == 0) {
					params->out_endian = MKESA_BIG_ENDIAN;

				} else {
					fprintf(stderr, "Error: Invalid output endianess: %s.\
(Valid: 'little' or 'big')\nTry --help for details.\n", optarg);
					exit(EXIT_FAILURE);
				}
				break;

			case 's':
				/* Input bit width switch */
				if(strncmp(optarg, "32", 2) == 0) {
					params->in_uint_size = fid_UINTSIZE_32;

				} else if(strncmp(optarg, "64", 2) == 0) {
					params->in_uint_size = fid_UINTSIZE_64;

				} else {
					fprintf(stderr, "Error: Invalid input integer width: %s.\
(Valid: '32' or '64')\nTry --help for details.\n", optarg);
					exit(EXIT_FAILURE);
				}
				break;

			case 'S':
				/* Output bit width switch */
				if(strncmp(optarg, "32", 2) == 0) {
					params->out_uint_size = fid_UINTSIZE_32;

				} else if(strncmp(optarg, "64", 2) == 0) {
					params->out_uint_size = fid_UINTSIZE_64;

				} else {
					fprintf(stderr, "Error: Invalid output integer width: %s.\
(Valid: '32' or '64')\nTry --help for details.\n", optarg);
					exit(EXIT_FAILURE);
				}
				break;

			case '?':
			case ':':
				/* Unknown argument, getopt will print its own error message
				as well. */
				fprintf(stderr, "Try --help for usage information\n");
				exit(EXIT_FAILURE);
				break;

			default:
				/* This should never happen. */
				fprintf(stderr, "Error: Internal error while parsing arguments.\n");
				exit(EXIT_FAILURE);
				break;
		}
	}

	if(optind < argc)
	{
		int multiple=(argc-optind > 1);

		fprintf(stderr,"Error: Invalid option%s:",multiple?"s":"");
		if(multiple)
		{
			fputc('\n',stderr);
			while(optind < argc) fprintf(stderr,"  %s\n",argv[optind++]);
		}
		else fprintf(stderr," %s\n",argv[optind]);
		return -1;
	}

	/* Make sure that the user specified some
	kind of input files. */
	if(params->input_file == NULL) {
		fprintf(stderr, "Error: No input file specified.\n\
Use the -i switch or try --help for details.\n");
		exit(EXIT_FAILURE);
	} 

	if(params->output_file == NULL) {
		fprintf(stderr, "Error: No output file specified.\n\
Use the -o switch or try --help for details.\n");
		exit(EXIT_FAILURE);
	} 
		
	/* Make sure the user specified the input UINT size. */
	if(params->in_uint_size == fid_UINTSIZE_UNDEF) {
		fprintf(stderr, "Error: No input integer width specified.\n\
Use the -s switch or try --help for details.\n");
		exit(EXIT_FAILURE);
	} 
	
	/* Make sure the user specified the input endianess. */
	if(params->in_endian == MKESA_UNDEF_ENDIAN) {
		fprintf(stderr, "Error: No input endianess specified.\n\
Use the -e switch or try --help for details.\n");
		exit(EXIT_FAILURE);
	} 

	if(params->out_endian == MKESA_UNDEF_ENDIAN) {
		/* If no output endianess was specified, use
		the input endianess. */
		params->out_endian = params->in_endian;
	}

	if(params->out_uint_size == fid_UINTSIZE_UNDEF) {
		/* If no output UINT size was specified, use
		the input size. */
		params->out_uint_size = params->in_uint_size;
	}
	return 0;
}

static inline void write_error(FILE* infile, FILE* outfile) {
	/* Print write error message. */
	perror("Fatal error: Could not write to output file.");

	/* Close the files. */
	fclose(infile);
	fclose(outfile);

	/* And exit... */
	exit(EXIT_FAILURE);
}



static inline void switch_endian_32(char *const buf) {
	/* Switch between little and big endian, 32 bit data */
	SWAP_CHAR(buf[3], buf[0]);
	SWAP_CHAR(buf[2], buf[1]);
	return;
}

static inline void switch_endian_64(char *const buf) {
	/* Switch between little and big endian, 64 bit data */
	SWAP_CHAR(buf[7], buf[0]);
	SWAP_CHAR(buf[6], buf[1]);
	SWAP_CHAR(buf[5], buf[2]);
	SWAP_CHAR(buf[4], buf[3]);
}

int main(int argc, char* argv[]) {
	convesa_params_t params;
	FILE *infile, *outfile;
	char buf[8];
	size_t uint_bytes, bytes_read;
	unsigned char switch_endian = 0, scale = 0;

	/* Parse the command line argumets */
	parse_params(argc, argv, &params);

	/* Get the input UINT size in bytes */
	if(params.in_uint_size == fid_UINTSIZE_32) {
		uint_bytes = 4;
	} else {
		uint_bytes = 8;
	}

	if(params.in_endian != params.out_endian) {
		/* Check if we have to switch endinaess */
		switch_endian = 1;
	}

	if(params.in_uint_size != params.out_uint_size) {
		/* Check if we have to switch bit width */
		scale = 1;
	}

	if(params.verbose) {
		printf("Opening input file %s.\n", params.input_file);
	}
	/* Try to open the input file */
	infile = fopen(params.input_file, "r");

	if(infile == NULL) {
		perror("Fatal error: could not open input file for reading.");
		exit(EXIT_FAILURE);
	}

	if(params.verbose) {
		printf("Opening output file %s.\n", params.output_file);
	}
	/* Try to open the output file */
	outfile = fopen(params.output_file, "w");

	if(outfile == NULL) {
		perror("Fatal error: could not open output file for reading.");
		fclose(infile);
		exit(EXIT_FAILURE);
	}

	while(!feof(infile)) {
		/* Read another integer from the input file */
		bytes_read = fread(buf, 1, uint_bytes, infile);

		if(bytes_read != uint_bytes) {
			/* Having less bytes read than specified indicates
			a read error or EOF */
			if(ferror(infile)) {
				/* Check for read errors */
				perror("Fatal error: Read error while reading from input file.");

				fclose(infile);
				fclose(outfile);

				exit(EXIT_FAILURE);
			} else if(bytes_read != 0) {
					/* If the final read didn't read enough bytes for a whole integer,
					warn the user that we ignored the 'dangling' (remaining) bytes. */
					printf("Warning: Premature end of input, (input was not a\n\
multiple of %lu), %lu dangling bytes ignored.\n", 
(unsigned long) uint_bytes, (unsigned long) bytes_read);
			}

			/* Leave loop on EOF */
			break;
		}

		if(switch_endian) {
			/* Switch the endianess of the integer if required */
			if(uint_bytes == 8) {
				switch_endian_64(buf);
			} else {
				switch_endian_32(buf);
			}
		}

		if(scale) {
			/* We have to convert from 32 to 64 bit or vice versa */
			if(uint_bytes == 4) {
				/* 32 -> 64 bit upscaling mode */
				/* Insert four \0 characters for little endian padding
				(32 -> 64 bit integer conversion). */
				if(params.out_endian == MKESA_BIG_ENDIAN) {
					if(fwrite("\0\0\0\0", 1, 4, outfile) != 4) {
						write_error(infile, outfile);
					}
				}

				/* Write the (converted) integer */
				if(fwrite(buf, 1, uint_bytes, outfile) != uint_bytes) {
						write_error(infile, outfile);
				}

				/* Append four \0 characters for big endian padding
				(32 -> 64 bit integer conversion). */
				if(params.out_endian == MKESA_LITTLE_ENDIAN) {
					if(fwrite("\0\0\0\0", 1, 4, outfile) != 4) {
						write_error(infile, outfile);
					}
				}
			} else {
				/* 64 -> 32 bit downscaling mode */
				if(params.out_endian == MKESA_LITTLE_ENDIAN) {
					/* Write the least significant half of
					the input */
					if(fwrite(buf, 1, 4, outfile) != 4) {
						write_error(infile, outfile);
					}

					/* Warn the user about information that was lost
					while downscaling */
					if(*((fid_Uint32*)(buf + 4)) != 0) {
						printf(
"Warning: Information lost while converting from 64 to 32 bit.\n");
					}
				} else {
					/* Write the least significant half of
					the input */
					if(fwrite(buf + 4, 1, 4, outfile) != 4) {
						write_error(infile, outfile);
					}

					/* Warn the user about information that was lost
					while downscaling */
					if(*((fid_Uint32*)buf) != 0) {
						printf(
"Warning: Information lost while converting from 64 to 32 bit.\n");
					}
				}
			}
		} else {
			/* No scaling required, simply write the output */
			if(fwrite(buf, 1, uint_bytes, outfile) != uint_bytes) {
				write_error(infile, outfile);
			}
		}
	}

	if(params.verbose) {
		printf("Done, cleaning up.\n");
	}

	/* Close the input/output files */
	fclose(infile);
	fclose(outfile);

	return 0;
}
