/*
	Copyright (C) 2008, 2009 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef MULTITHREADED
#include <pthread.h>
#endif
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif

#include "common.h"
#include "fid.h"
#include "util.h"

#define LLV_BLOCK_SIZE 2048

extern void qsort_rank(UINT *a, size_t n, const UINT* sufarray);

static void error_dump(fid_Error* error, const char* errmsg) {

	if(errmsg != NULL) {
		fprintf(stderr, "%s", errmsg);
	}
  fid_error_dump(error, stderr);
	fid_error_free(error);
}



/*!
 * \brief Loads the alphabet specified in a set of parameters.
 *
 * If an alphabet file is specified in the mkesa_params_t structure
 * the alphabet will be read from the file. If no file is given,
 * the specified built-in alphabet will be used.
 * \param alpha A pointer to an unitialized alphabet.
 * \param params The application parameters as specified by the user.
 * \returns 0 on success, error code otherwise.
 */
int get_alphabet(fid_Alphabet* alpha, const mkesa_params_t* params) {
	fid_Error error;
	int retcode = 0;

	/* No alphabet file given -> use built-un alphabet */
	if(params->al1_file == NULL) {
		if(params->verbose > 1) {
			printf("Generating standard alphabet.\n"); 
		}
		/* Load built-in alphabet */
		fid_alphabet_init_standard(alpha, params->builtin_alphabet);
		return 0;

	} else {
		if(params->verbose) {
			printf("Attempting to load alphabet from file \"%s\".\n",
					params->al1_file);
		}

		/* Initialize error structure */
		fid_error_init_default(&error);

		/* Load alphabet from file */
		retcode = fid_alphabet_init_from_specfile(alpha, params->al1_file,
				&error);

		/* Checkl for errors */
		if(retcode != 0){
			error_dump(&error, "Error while loading alphabets:\n");
			return retcode;

		} else {
			if(params->verbose) {
				printf("Alphabet loaded successfully.\n");
			}
		}

		/* Clean up the error structure */
		fid_error_free(&error);
	}

	return retcode;
}


/*!
 * \brief Loads a transformed input sequence with padding.
 *
 * The input sequence is read from the project specific
 * TIS file and stored to the tisfile pointer of the esa
 * mkesa_params_t parameter after being padded with 
 * the specified number of separator symbols.
 * The padding is used by various sorting algoritms that would have
 * to perform explicit "end-of-string" checks while comparing
 * two suffixes. The special properties of separators ensure that
 * a comparison between two suffixes always terminates if a special
 * character is encountered. Therefore by padding the input sequence
 * with special characters we can ensure that a comparisson between
 * two suffixes always terminates without the need for explicit 
 * "end-of-string" checks.
 * \param esa Suffix array structure, a pointer to the
 * loaded input sequence will be stored within this structure.
 * \param params Program parameters, contains the project name
 * that is used for determining the TIS file filename.
 * \param padding Number of fid_SEPERATOR symbols appended 
 * to the loaded input sequence.
 * \param error Pointer to a fid_Error structure, used
 * for passing error information to the caller if appropriate.
 * \returns 0 on success, non-zero error code on error.
 */
static int load_tis_padded(fid_Suffixarray *const esa,
		const mkesa_params_t* params, UINT padding,
		fid_Error *const error) {

	assert(esa != NULL);
	assert(params != NULL);
	assert(error != NULL);

	fid_Symbol* tismem;
	UINT seqlen, i;
	int retcode = 0;

	char* tisfile =
			fid_filename_create(params->prj_name, "tis", error);

	if(tisfile == NULL) {
		return -1;
	}

	/* Try to load existing tables for update mode */
	retcode = fid_file_map(&(esa->sequences.tisfile),
			tisfile, 0, 1, error);

	free(tisfile);

	if(retcode != 0) {
		return retcode;
	}

	seqlen = esa->sequences.tisfile.occupied;
	assert(seqlen > 0);

	esa->sequences.total_length.VU = seqlen;

	/* Allocate memory for the TIS data */
	tismem = malloc_safe(sizeof(fid_Symbol) * (seqlen + padding));
	assert(tismem != NULL);
	
	/* Copy TIS to the larger space */
	memcpy(tismem, esa->sequences.tisfile.content, seqlen);

	/* Use seperator symbols for padding */
	for(i = 0; i < padding; i++) {
		tismem[seqlen + i] = fid_SEPARATOR;
	}

	/* Unmap the original file... */
	fid_file_unmap(&(esa->sequences.tisfile));

	/* ...and replace it with the new fake file */
	fid_file_fake(&(esa->sequences.tisfile), tismem,
			sizeof(fid_Symbol) * (seqlen + padding));

	/* Mark file as full */
	esa->sequences.tisfile.occupied =
			esa->sequences.tisfile.allocated;

	return retcode;
}

/*!
 * \brief Creates the necessary tables from the input file and
 * 	maps them to memory.
 *
 * 	First the alphabet will be loaded if necessary and written
 * 	into a project-specific alphabet file.
 * 	Then the input sequence is read to memory from the file 
 *  specified within \p params and transformed using the alphabet.
 * 	The suffix array file is created and initialized as well.
 *
 * \param alpha The alphabet that should be used for all
 * 	transformations. Can be NULL, in which case the alphabet
 * 	specified by the application parameters will be loaded.
 * \param esa Pointer to a uninitialized fid_Suffixarray structure.
 * 	Pointers to the mapped sequences will be written to this
 * 	structure after the structure was initialized.
 * \param params The application parameters as specified by the user.
 * \returns 0 on success, non-zero error code on error.
 */
int build_tables(fid_Alphabet* alpha,
		fid_Suffixarray* esa, mkesa_params_t* params) {
	assert(esa != NULL);
	assert(params != NULL);

	int retcode;
	void* sufmem = NULL;

	fid_Error error;
	fid_Alphabet new_alpha;
	fid_Projectfile prjfile;

	/* Initialize error structure */
	fid_error_init_default(&error);

	/* Initialize suffix array structure. */
	fid_suffixarray_init(esa, params->uint_size);

	/* Try to load the alphabet if none was given. */
	if(!params->update_mode) {
		if(alpha == NULL) {
			if(get_alphabet(&new_alpha, params) != 0) {
				return -1;
			}
			alpha = &new_alpha;
		}

		if(params->verbose) {
			printf("Storing alphabet to \"%s.al1\"\n", params->prj_name);
		}

		/* Write the alphabet to a new project-specific alphabet file. */
		retcode = fid_alphabet_write_to_file(alpha, params->prj_name, &error);

		if(retcode != 0) {
			error_dump(&error, "Error while saving alphabet.\n");
			return retcode;
		}

		if(params->verbose) {
			printf("\nAttempting to transform input sequences.\n");
		}
	
		/* Generate transformed input sequence from input data file 
		 * using the specified alphabet and UINT size. */
		fid_Tablerequest request = fid_TABLE_TIS | fid_TABLE_DES;
		if(params->build_ois) request |= fid_TABLE_OIS;
		retcode = fid_sequences_parse_from_file_to_memory(params->data_file,
				fid_FORMAT_UNDEF, request, &(esa->sequences),
				alpha, 2, params->uint_size, &(params->input_file_size), &error);

		if(retcode != 0) {
			error_dump(&error, "Error while transforming sequences.\n");
			return retcode;
	
		} else {
			if(params->verbose) {
				printf("Sequences transformed successfully.\n");
			}
		}
	} else {
		/* Update mode, only create the necessary tables
		 * and avoid re-creating existing ones. */

		if(params->verbose) {
			printf("Loading project file for existing project.\n");
		}

		/* Check and load project file */
	 	retcode = fid_projectfile_parse_from_file(&prjfile,
			params->prj_name, &error);

		if(retcode != 0) {
			/* The project file is considered an integral part of a given project.
			 * Therefore not being able to load it is considered to be a fatal error. */
			error_dump(&error, "Error while loading project file for update mode.\n");

			fprintf(stderr,
"\nIf you want to start from scratch with a new project file please\n\
specify a data file using the -d command or see --help for details.\n\n");
			return retcode;
		}

		/* Read number of large lcp values from project file */
		esa->num_of_large_lcps.VU = prjfile.large_lcp_values;

		/* Check if the current UINT size matches the one specified
			 in the project file. Error & abort if not. */
		if(params->uint_size != prjfile.uisize) {
			fprintf(stderr, 
"\nError: The integer size specified in the project file (%u bit) doesn't match\n\
the integer size used by this binary (%u bit).\n\
You can either change the integer size used using -w or re-create the project\n\
from scratch with the right integer size using the -d command.\n\
See --help for details.\n\n", prjfile.integer_size_in_bits, 
					((params->uint_size == fid_UINTSIZE_32) ? 32 : 64));

			fid_projectfile_free(&prjfile);
			fid_error_free(&error);
			return -1;
		}

		if(params->have_tis)
		{
			/* Load TIS file with padding */
			if(params->verbose) {
				printf("Loading existing TIS file.\n");
			}

			retcode = load_tis_padded(esa, params, 2, &error);

			if(retcode != 0) {
				error_dump(&error, "Error while loading TIS file for update mode.\n");
				return retcode;
			}
		}
		else esa->sequences.total_length.VU = prjfile.totallength;

		/* Clean up the project file */
		fid_projectfile_free(&prjfile);

	}

	/* Free error structure */
	fid_error_free(&error);
	
	/* Attempt to allocate memory for the suffix array file. */
	sufmem = malloc_safe(sizeof(UINT) * (esa->sequences.total_length.VU + 1));
	assert(sufmem != NULL);

	/* Add fake memory for the suffix array file */
	fid_file_fake(&(esa->suffile), sufmem,
			sizeof(UINT) * (esa->sequences.total_length.VU + 1));
	esa->suftab.VU = (const UINT *)esa->suffile.content;

	/* Assign alphabet */
	esa->alpha = *alpha;

	/* Mark the entire file as occupied since all available space
	  will be used later on. */
	esa->suffile.occupied = esa->suffile.allocated;

	return 0;
}

/*!
 * \brief Writes a single mapped table to an actual file.
 *
 * The actual file name is generated by concatenating the 
 * base name with a dot (".") and the file extension \p ext.
 * If the \p verbose flag is set then a status message is printed
 * to stdout upon successful write. Error messages are written
 * to stderr should an error occur.
 * \param base The base name of the output file.
 * \param ext The file extension of the output file.
 * \param table The table that should be written.
 * \param verbose Verbose flag.
 * \returns 0 upon success, error otherwise.
 */
static int write_fake_table(const char* base,
		const char* ext, const fid_Mappedfile* table, const int verbose) {
	assert(base != NULL);
	assert(ext != NULL);
	assert(table != NULL);

	int retcode;
	fid_Error error;

	char* filename = NULL;

	/* Initialize error structure */
	fid_error_init_default(&error);

	/* Append file extension to base name */
	filename = fid_filename_create(base, ext, &error);

	if(filename == NULL) {
		error_dump(&error, "Internal error while generating file name.\n");
		fid_error_free(&error);
		return -1;
	}

	fid_error_free(&error);

	/* Dump fake mapped file to file. */
	retcode = fid_file_dump_to_file(table, filename, &error);

	if(retcode != 0) {
		error_dump(&error, "Error while writing file.\n");
		return retcode;
	} else {
		if(verbose) {
			printf("File written successfully.\n");
		}
	}

	/* Cleanup */
	fid_error_free(&error);
	free(filename);

	return retcode;
}

/*!
 * \brief Build and write inverse suffix array table.
 *
 * This function will build an inverse suffix array
 * and write it to a file. An inverse suffix array
 * is an array that maps the starting prosition of
 * each suffix of the input sequence to its 
 * position in the suffix array.
 * The output will be written to PROJECT_NAME.sti, where
 * PROJECT_NAME is the project name as specified by the
 * mkesa_params_t data structure.
 * Appropriate error messages will be written to stderr
 * if an error is encountered.
 * Info messages will be written to stdout if in verbose mode.
 * \param esa The suffix array used as the source for the
 * inverse suffix array.
 * \param params Program parameters, contains the project
 * name used as the output file name.
 * \returns 0 on success, non-zero error code on error.
 */
static int write_sti_tab(fid_Suffixarray *esa,
		const mkesa_params_t *params) {
	UINT i, *stimem, *suftab;
	int retcode = 0;

	if(params->verbose) {
		printf("Generating inverse suffix table: ");
		fflush(stdout);
	}

	suftab = (UINT*) esa->suffile.content;

	/* Allocate memory for the inverse table */
	stimem = malloc_safe(sizeof(UINT) * (esa->sequences.total_length.VU + 1));
	assert(stimem != NULL);

	/* Generate fake mapped file using the memory allocated above */
	fid_file_fake(&(esa->stifile), (void*) stimem,
			sizeof(UINT) * (esa->sequences.total_length.VU + 1));

	/* Manually mark the entire memory as used */
	esa->stifile.occupied = esa->stifile.allocated;

	/* Simple loop that does the actual work of generating
	 * the inverse suffix array. */
	for(i = 0; i < esa->sequences.total_length.VU + 1; i++) {
		assert(suftab[i] <= esa->sequences.total_length.VU + 1);
		stimem[suftab[i]] = i;
	}

	/* Write the table to a file */
	retcode = write_fake_table(params->prj_name, "sti", &(esa->stifile),
			params->verbose);

	/* Unmap the fake file, this also frees the allocated  memory */
	fid_file_unmap(&(esa->stifile));

	return retcode;
}

/*!
 * \brief Precompute inverse suffix array and store it in the lcp-table.
 *
 * The array values are stored in a special order so that they can be
 * overwritten while the lcp-table is computed. This is necessary to reduce the
 * space consumption from 13n (Kasai 2001) to 9n bytes.
 *
 * \param esa The enhanced suffix array the lcp-table should be computed for.
 * \param lcptable Intermediate space which will become the lcp-table. This
 * function stores the inverse suffix array in this array.
 *
 * \returns The index of the first entry in \p lcptable for the main lcp-table
 * computation algorithm to start with.
 *
 * \note This function corresponds to Sa2RankNext() in Manzini 2004. It is a
 * bit more complicated then the version presented in the paper, however, since
 * separators and wildcards are encoded by special symbols, and these are not
 * equal to any other symbol, even not to themselves. The corresponding
 * suffixes are stored in the suffix array in the order as they appear in the
 * text, so they are not sorted lexicographically and thus destroy the suffix
 * array properties. Hence, the Burrows-Wheeler-transformation does not work
 * reliably when computed on-the-fly as done by this algorithm. The values for
 * the special symbols are all known after rank construction, but as a
 * consequence of the non-comparability of special symbols, they must be sorted
 * in a separate step to put them into the correct place for the in-place
 * lcp-table computation. Since there should be only few special symbols in a
 * database, it seems reasonable to make this minor sacrifice instead of
 * changing the whole binary encoding of the suffix array (and thus, all
 * algorithms working with them).
 */
static UINT rank_next(const fid_Suffixarray *esa, UINT *lcptable) {
  UINT count[UCHAR_MAX + 2];
  UINT seqlen = esa->sequences.total_length.VU;
  UINT i, retval = 0;
	UINT *suftab = (UINT*) esa->suffile.content;
  fid_Symbol sym;

  /* compute cumulated histogram */
  memset(count, 0, sizeof(UINT) * (UCHAR_MAX + 2));
  for(i = 0; i < seqlen; i++) {
    count[esa->sequences.tisfile.content[i] + 1]++;
  }
  for(i = 1; i <= UCHAR_MAX + 1; i++) {
    count[i] += count[i - 1];
  }

#ifdef DEBUG
  memset(lcptable, 0xff, (size_t)(seqlen + 1) * sizeof(UINT));
#endif /* DEBUG */

  /* treat first suffix special if possible (otherwise the value is not needed
   * anyway) */
  if(suftab[0] > 1) {
    sym = esa->sequences.tisfile.content[suftab[0] - 1];
    if(fid_REGULARSYMBOL(sym)) {
      lcptable[count[(size_t)sym]++] = 0;
    } else {
      lcptable[count[(size_t)fid_WILDCARD]++] = 0;
    }
    i = 1;
  }
  else
  {
    i = 0;
  }

  /* Compute ranks. */
  for(/* nothing */; i < seqlen; i++) {   
    if(suftab[i] > 0) {
      sym = esa->sequences.tisfile.content[suftab[i] - 1];
      if(fid_REGULARSYMBOL(sym)) { 
        assert(lcptable[count[(size_t)sym]] == ~(UINT)0);
        lcptable[count[(size_t)sym]++] = i;
      } else {
        assert(lcptable[count[(size_t)fid_WILDCARD]] == ~(UINT)0);
        lcptable[count[(size_t)fid_WILDCARD]++] = i;
      }
    } else {
      retval = i;
    }
  }

	/* sort ranks for special symbols */
	if(seqlen > count[fid_WILDCARD - 1])
	{
		size_t len=(fid_REGULARSYMBOL(esa->sequences.tisfile.content[seqlen - 1])
			?seqlen - count[fid_WILDCARD - 1]
			:seqlen - count[fid_WILDCARD - 1] - 1);
		qsort_rank(&lcptable[count[fid_WILDCARD - 1]], len, suftab);
	}

  return retval;
}

/*!
 * \brief Computes and writes skip table from suffix array.
 *
 * \param esa The enhanced suffix array that the skip table will
 * be based on.
 * \param params Various application parameters.
 * 
 * \returns 0 on success, libfid error code on error.
 */
static int write_skiptab(fid_Suffixarray *esa,
		const mkesa_params_t* params) /*{{{*/
{
  UINT i, j;
  UINT ilcp, jlcp, seqlen;
	UINT *skpfile;
	int retcode;

	if(params->verbose) {
		printf("Generating skip table: ");
		fflush(stdout);
	}

  seqlen = esa->sequences.total_length.VU;

	/* Get memory for the skip table */
	skpfile = malloc(sizeof(UINT) * (seqlen + 1));
	
	/* Initialize fake file and set occupied space to 0 */
	fid_file_fake(&(esa->skpfile), (void*) skpfile,
			sizeof(UINT) * (seqlen + 1));
	esa->skpfile.occupied = 0; 

  for(i=0; i < seqlen; i++) {
    fid_LCP(ilcp, esa, i);
    if(ilcp > 0) {
      for(j=i+1; j < seqlen; j++) {
        fid_LCP(jlcp, esa, j);
        if(ilcp > jlcp) {
          break;
        }
      }
      fid_MAPPEDFILE_APPEND_UNSAFE(&esa->skpfile, UINT, j-1);
    } else {
      fid_MAPPEDFILE_APPEND_UNSAFE(&esa->skpfile, UINT, esa->sequences.total_length.VU);
    }
  }
  
  fid_MAPPEDFILE_APPEND_UNSAFE(&esa->skpfile, UINT,esa->sequences.total_length.VU);

	/* Write skip table to file */
	retcode = write_fake_table(params->prj_name, "skp", &(esa->skpfile),
			params->verbose);

	/* Clean up the allocated memory */
	fid_file_unmap(&(esa->skpfile));

  return retcode;
}

/*!
 * \brief Compute lcp-table with algorithm lcp9 due to Manzini 2004, a more
 * space efficient version of Kasai et al. 2001.
 * Also calls the function for creating the skip table if requested.
 * This is done in one piece as to save memory / address space through 
 * just-in-time resource allocation / clean-up.
 * 
 * \param esa The enhanced suffix array to be computed.
 * \param params Data structure providing various application parameters.
 * 
 * \returns 0 on success, -1 on error.
 */
static int write_lcp_skp_tab(fid_Suffixarray *esa,
		const mkesa_params_t* params) {
	UINT *lcptable, *llvtable, *suftab;
  UINT seqlen, lcpvalue, i, j, k, nextk;
	char* filename;

	fid_Error error;
	fid_error_init_default(&error);

  fid_Symbol sym1, sym2;

	int retcode = 0;

	fid_Uint64 llv_length;

	if(!params->update_mode ||
		 !test_file_readable(params->prj_name, "lcp") ||
		 !((test_file_readable(params->prj_name, "llv") &&
		 	 (esa->num_of_large_lcps.VU != 0)))) {

		/* Re-set the large lcp value (llv) count */
		esa->num_of_large_lcps.VU = 0;

		if(params->verbose) {
			printf("Generating LCP & LLV tables: ");
			fflush(stdout);
		}

		suftab = (UINT*) esa->suffile.content;
  	seqlen = esa->sequences.total_length.VU;
	
	
		/* Allocate memory for the main lcp table */
		lcptable = (UINT *) malloc_safe((size_t)(seqlen + 1) * sizeof(UINT));
		assert(lcptable != NULL);

		/* Allocate memory for the llv table, start with enough
			 memory for a few entries. */
		llv_length = LLV_BLOCK_SIZE;
		llvtable = (UINT *) malloc_safe(llv_length * sizeof(UINT));
		assert(llvtable != NULL);

		/* Generate fake mapped file using the memory allocated above */
		fid_file_fake(&(esa->lcpfile), (void*) lcptable,
				sizeof(UINT) * (seqlen + 1));

		/* Mark memory as free */
		esa->lcpfile.occupied = 0; 

  	k = rank_next(esa, lcptable);

  	/* compute lcp-table */
  	lcpvalue = 0;

  	for(i = 0; i < seqlen; i++) {
    	nextk = lcptable[k];
    	if(k > 0) {
      	assert(k-1 <= seqlen);
      	j = suftab[k-1];
	
      	while(i + lcpvalue < seqlen && j + lcpvalue < seqlen &&
            	(sym1 = esa->sequences.tisfile.content[i + lcpvalue]) ==
            	(sym2 = esa->sequences.tisfile.content[j + lcpvalue]) &&
           	fid_REGULARSYMBOL(sym1) && fid_REGULARSYMBOL(sym2)) {
        	lcpvalue++;
	      }
	      lcptable[k] = lcpvalue;
	    }
	    else {
	      lcptable[k] = 0;
	    }
	
	    if(lcpvalue > 0) {
	      lcpvalue--;
	    }
	
	    k = nextk;
	  }
	 
		/* If everything went OK, then the first suffix should have
			 a LCP of zero since it has no left neightbour */
	  assert(lcptable[0] == 0);
		/* The lcp of the last (read: zero-length) suffix is
			 always zero */
	  lcptable[seqlen] = 0;
	
	  for(i=0; i <= seqlen; i++) {
			/* Regular, e.g. small LCP, append to the LCP file */
	    if(lcptable[i] < UCHAR_MAX) {
	      fid_MAPPEDFILE_APPEND_UNSAFE(&esa->lcpfile, unsigned char, lcptable[i]);
	    }
	    else {
				/* Long LCP, append to LLV file */
	
				/* If we are out of memory ... */
				if((esa->num_of_large_lcps.VU * 2) >= llv_length) {
					/* ... allocate more memory for the llv table.
						Allocate in multi-byte chunks, this minimizes the number of
						system calls while causing an acceptable overhead */
					llv_length += LLV_BLOCK_SIZE;
					llvtable = realloc_safe(llvtable, llv_length * sizeof(UINT));
					assert(llvtable != NULL);
				}
	
				/* Mark this LCP as a long LCP in the reqular LCP file */
	      fid_MAPPEDFILE_APPEND_UNSAFE(&esa->lcpfile, unsigned char, UCHAR_MAX);
	
				/* Write long LCP position and value to llv file */
				llvtable[esa->num_of_large_lcps.VU * 2] = i;
				llvtable[(esa->num_of_large_lcps.VU * 2) + 1] = lcptable[i];
	
				/* Increase LLV counter */
				esa->num_of_large_lcps.VU++;
	    }
	  }
	
		/* Write the table to a file */
		write_fake_table(params->prj_name, "lcp", &(esa->lcpfile),
				params->verbose);
	
		if(esa->num_of_large_lcps.VU != 0) {
			/* Create and write LLV fake file. We cannot create this file at the
				 beginning since the actual size of the LLV file isn't known up to this
				 point */
			fid_file_fake(&(esa->llvfile), (void*) llvtable,
				esa->num_of_large_lcps.VU * 2 * sizeof(UINT));
	
			write_fake_table(params->prj_name, "llv", &(esa->llvfile),
				0);
	
		} else if(params->verbose) {
			printf("No long LCPs, llv file not written.\n");
			free(llvtable);
		}
	} else if(params->build_skp &&
		(!test_file_readable(params->prj_name, "skp"))) {
		/* Load existing (old) LCP/LLV files
			 if, and only if, we're a) in update mode and b) we did not
			 generate  new LCP/LLV files right now and
			 c) the skip table does not yet exist. */
		if(params->verbose) {
			printf("Loading existing LCP and LLV files.\n");
		}

		fid_error_free(&error);

		/* Attempt to map existing LCP file
			 for use in skip table generation. */
		filename = fid_filename_create(params->prj_name,
				"lcp", NULL);
		retcode = fid_file_map(&(esa->lcpfile), filename,
			0, 1, &error);

		if(retcode != 0) {
			error_dump(&error, "Could not map existing LCP file.\n");
			return retcode;
		}
			
		fid_error_free(&error);

		if(esa->num_of_large_lcps.VU > 0) {
			/* Only attempt to map the LLV file if it should exist,
			 * e.g. the large LCP count is greater than 0. */

			filename = fid_filename_create(params->prj_name,
					"llv", NULL);
			retcode = fid_file_map(&(esa->llvfile), filename,
				0, 1, &error);

			if(retcode != 0) {
				error_dump(&error, "Could not map existing LLV file.\n");
				return retcode;
			}
		}
	} else {
		if(params->verbose) {
			printf("LCP/LLV files alreay exist, not modified.\n");
		}
	}
	
	if(params->build_skp) {
		if((!params->update_mode) ||
			(!test_file_readable(params->prj_name, "skp"))) {
			/* build skip table if it was requested and we're
			 * either a) in create mode or b) in update mode and
			 * the skip table does not yet exist. */
			write_skiptab(esa, params);
		} else if(params->verbose) {
			printf("Skip table file already exists, not written.\n");
		}
	}

	/* Unmap the fake files, this also frees the allocated memory */
	fid_file_unmap(&(esa->lcpfile));

	fid_file_unmap(&(esa->llvfile));

	fid_error_free(&error);

	return 0;
}


/*!
 * \brief Create a project file from known information and write it
 * to the appropriate file on disk.
 * \param esa Suffix array structure that the projet file will be
 * generated from.
 * \param params Program parameters. The actual file name
 * of the project file is derived by appending the ".prj" file
 * extension to the project name specified within this structure.
 * \returns 0 on success, non-zero error code on error.
 */
static int create_and_write_prjfile(const fid_Suffixarray* esa,
		const mkesa_params_t* params) {
	int retcode;

	fid_Projectfile prjfile;

	fid_Error error;
	fid_error_init_default(&error);

	fid_DBfiles dbfiles;
	fid_DYNARRAY_INIT(&dbfiles,fid_DBfileinfo);

	if(params->verbose > 1) {
		printf("Writing project file: ");
	}

	/* Create dbfiles */
	retcode = fid_dbfiles_append_copy(&dbfiles, params->data_file,
			params->input_file_size, esa->sequences.total_length.VU, &error);

	if(retcode != 0) {
		error_dump(&error, "Error while creating database files.\n");
		return retcode;
	}

	/* Initialize project file */
	retcode = fid_projectfile_init_from_suffixarray(&prjfile, &dbfiles,
			params->prj_name, esa, &error);

	if(retcode != 0) {
		error_dump(&error, "Error while creating project file.\n");
		return retcode;
	}

	/* Write project file to disk */
	retcode = fid_projectfile_write(&prjfile, &error);

	if(retcode != 0) {
		error_dump(&error, "Error while writing project file.\n");
		return retcode;
	}

	/* Clean-up */
	fid_error_free(&error);
	fid_projectfile_free(&prjfile);
	fid_dbfiles_free(&dbfiles);

	if(params->verbose > 1) {
		printf("File written successfully\n");
	}

	return 0;
}
/*! 
 * \brief Write all relevant fake mapped files to actual files.
 *
 * The suffix array, transformed input sequence and sequence
 * separator positions are written to the 
 * basename + ".suf", ".tis" and ".ssp" files. Also prints status
 * messages if the verbose flag of the mkesa_params_t structure
 * is set. Upon error an error message is printed and the function
 * returns with the error code.
 * \param esa Suffix array structure containing tables.
 * \param params Command line parameters.
 * \returns 0 upon success, error code otherwise.
 */
int write_and_free_tables(fid_Suffixarray* esa, const mkesa_params_t* params) {
	assert(params != NULL);
	assert(esa != NULL);

	fid_Error error;
	fid_Projectfile prjfile;
	int retcode;

	/* Initialize error structure */

	if(params->verbose) {
		printf("\nWriting output files...\n");
	}

	if(params->write_suf) {
		if(params->verbose) {
			printf("Writing suffix array file: ");
			fflush(stdout);
		}

		/* Writing suffix array */
		retcode = write_fake_table(params->prj_name, "suf", &(esa->suffile),
				params->verbose);

		if(retcode != 0) {
			return retcode;
		}
	} 
	
	if(!params->update_mode) {
		/* All these files are only written in create mode */

		if(params->verbose) {
			printf("Writing transformed input sequence: ");
			fflush(stdout);
		}

		/* Writing transformed input sequence */
		retcode = write_fake_table(params->prj_name, "tis", &(esa->sequences.tisfile),
				params->verbose);

		if(retcode != 0) {
			return retcode;
		}

		if(params->build_ois)
		{
			if(params->verbose) {
				printf("Writing original input sequence: ");
				fflush(stdout);
			}

			/* Writing original input sequence */
			retcode = write_fake_table(params->prj_name, "ois", &(esa->sequences.oisfile),
																 params->verbose);

			if(retcode != 0) {
				return retcode;
			}

			fid_file_unmap(&esa->sequences.oisfile);
		}

		/* Sequence descriptions */
		if(params->verbose) {
			printf("Writing sequence descriptions: ");
			fflush(stdout);
		}
		if(esa->sequences.desfile.content != NULL) {
			retcode = write_fake_table(params->prj_name, "des",
				&(esa->sequences.desfile), params->verbose);

			if(retcode != 0) {
				return retcode;
			}
		} else {
			printf("No sequence descriptions, file not written.\n");
		}


		if(params->verbose) {
			printf("Writing sequence description seperators: ");
			fflush(stdout);
		}

		/* Write description seperators */
		if(esa->sequences.sdsfile.content != NULL) {
			retcode = write_fake_table(params->prj_name, "sds",
				&(esa->sequences.sdsfile), params->verbose);

			if(retcode != 0) {
				return retcode;
			}
		} else {
			printf("No sequence description seperators, file not written.\n");
		}

		if(params->verbose) {
			printf("Writing separator positions: ");
			fflush(stdout);
		}

		/* Writing SSP file */
		if(esa->sequences.sspfile.content != NULL) {
			retcode = write_fake_table(params->prj_name, "ssp",
					&(esa->sequences.sspfile), params->verbose);

			if(retcode != 0) {
				return retcode;
			}
		} else {
			if(params->verbose) {
				printf("No seperators, file not written.\n");
			}
		}
	}

	/* Write inverse suffix table */
	if(params->build_lcp) {
		retcode = write_lcp_skp_tab(esa, params);

		if(retcode != 0) {
			return retcode;
		}
	}

	/* Update project file with large LCP values */
	if(params->update_mode) {
		if(params->verbose) {
			printf("Updating project file.\n");
		}

		fid_error_init_default(&error);

		/* Check and load project file */
 		retcode = fid_projectfile_parse_from_file(&prjfile,
			params->prj_name, &error);

		if(retcode != 0) {
			error_dump(&error, "Error while updating project file.\n");
			return retcode;
		}

		fid_error_free(&error);

		/* Update number of large LCPs */
		prjfile.large_lcp_values = esa->num_of_large_lcps.VU;

		retcode = fid_projectfile_write(&prjfile, &error);

		if(retcode != 0) {
			error_dump(&error, "Error while writing updated project file\n");
			return retcode;
		}

		fid_error_free(&error);
	} else {

		/* Write project file */
		retcode = create_and_write_prjfile(esa, params);

		if(retcode != 0) {
			return retcode;
		}
	}

	
	/* At this point we no longer need the input sequences,
		so let's clean them up to save address space */
	fid_file_unmap(&esa->sequences.tisfile);

	/* Write inverse suffix table */
	if(params->build_sti) {
		if((!params->update_mode) || 
					(!test_file_readable(params->prj_name, "sti"))) {
			retcode = write_sti_tab(esa, params);
		} else if(params->verbose) {
			printf("Inverse suffix array file already exists, not modified.\n");
		}
	}

	fid_suffixarray_free(esa);

	return 0;
	}
