/*
	Copyright (C) 2008, 2009 David Fleer, Robert Homann.

	This file is part of mkesa.

	Some parts of this file are based on the original Deep-Shallow
	implementation by Giovanni Manzini (from helped.c).

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef MULTITHREADED
#include <pthread.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "common.h"
#include "algo.h"
#include "blind.h"
#include "qsorts.h"
#include "induced.h"
#include "anchorinfo.h"

/* Marks a specific suffix for use with induction sort.
The method used for marking depends on the current use
of multithreading */
static inline void mark_suffix(UINT *const element,
	const sort_params_t* sp, UINT* bucket) {
#ifdef MULTITHREADED
	if(sp->threaded_sort) {
		/* Assertion for catching out of bounds exceptions */
		assert(element >= bucket && (size_t)(element - bucket) < (sp->marked_suffixes_size << 3));

		/* Mark suffix by using the thread-local
		bit vector */
		BITVEC_SET(sp->marked_suffixes, element - bucket);
	} else {
#endif
		/* Non-threaded sort, mark by setting the
		most significant bit of the suffix array entry */
		*element |= MSB_MASK;
#ifdef MULTITHREADED
	}
#endif
}

/* Tests if the suffix that begins "pos" symbols
 * before the *(ELEMENT) suffix is a member of the
 * suffixes that we want to sort. (Uses binary search).
 * If that is the case, then ELEMENT is marked by flipping
 * its most significant bit. */
static inline int test_and_mark(const UINT *const sufarray, const size_t n,
																const sort_params_t* sp, UINT *bucket,
																UINT *elemptr, signed long pos)
{
	UINT lower = 0, upper = n, mid;
	UINT elem=*elemptr;

	elem&=~MSB_MASK;

	while(lower < upper) {
		/* Slightly more complicated way of building
		 * mid but avoids the overflow that
		 * mid = (higher + lower) / 2 might cause */
		mid = lower + ((upper - lower) / 2);
		if((((sufarray[mid]&~MSB_MASK) + pos)&~MSB_MASK) < elem) {
			lower = mid + 1;
		} else {
			upper = mid;
		}
	}
	if((lower < n) && ((((sufarray[lower]&~MSB_MASK) + pos)&~MSB_MASK) == elem)) {
		/* Mark the suffix. The way this is done depends on wether
			 multithreading is enabled */
		mark_suffix(elemptr, sp, bucket);
		return 1;
	}

	return 0;
}

/*!
 * \brief Updates offset & anchor information with new data.
 *
 * The offset and anchor information contained within sort_params_t
 * is updated using information from the suffixes from the sufarray
 * array.
 * \param sufarray Array of suffixes.
 * \param n Number of suffixes in the array.
 * \param sp Sort parameter structure with pointers to the
 * 	offset / anchor arrays.
 */
void update_anchor(UINT *const sufarray, const size_t n,
	const sort_params_t* sp) {
	/* Anchor & offset arrays are a global resource, so we have to lock
		 mutex to use them if we're using multiple sorting threads */
	#ifdef MULTITHREADED
	if(sp->threaded_sort) {
		pthread_mutex_lock(sp->anchor_mutex);
	}
	#endif

	size_t i;
	UINT segment;
	for(i = 0; i < n; i++) {
		segment = sufarray[i] / sp->d;
		if((sufarray[i] % sp->d) < sp->offset[segment]) {
			sp->offset[segment] = (sufarray[i] % sp->d);
			sp->anchor[segment] = (sufarray + i);
			assert(*(sp->anchor[segment]) == segment*sp->d+sp->offset[segment]);
		}
	}

	#ifdef MULTITHREADED
	if(sp->threaded_sort) {
		pthread_mutex_unlock(sp->anchor_mutex);
	}
	#endif
}	

static void anchor_sort(UINT *sufarray, size_t n, const sort_params_t* sp,
                        Anchorinfo d, int backward)
{
	UINT *p, *q, *bucket, *bucket_end;
	UINT marked;
	UINT anchor_pos=
		(backward?sufarray[d.selected]-d.pos:sufarray[d.selected]+d.pos);
	UINT spos=backward?-d.pos:d.pos;

	assert(d.pos < 2U*sp->d);

	/* Find the bucket and bucket boundary, the bucket boundary
	 * is the beginninng of the next sub-bucket (This works since
	 * we are never dealing with the last sub-bucket in a bucket,
	 * which would be the special character bucket and is handled
	 * elsewhere.) */
	bucket    =sp->info->sub_buckets[sp->tp[anchor_pos]][sp->tp[anchor_pos+1]];
	bucket_end=sp->info->sub_buckets[sp->tp[anchor_pos]][sp->tp[anchor_pos+1]+1];

	/* We found a symbol pair belonging to a sorted sub-bucket,
	 * so now we have to sort the set of suffixes according to
	 * their offsets (Step 1-is). */
	qsort_offset(sufarray, n);

	/* Now we scan the suffixes to the left and 
	 * to the right of the first suffix inside the bucket.
	 * If a suffix (minus the position offset) belonging 
	 * to the input set is found, it is marked using the 
	 * most significant bit. We do this until all suffixes 
	 * in the input set have been found and marked. */
	p = sp->anchor[d.best];
	q = p + 1;

	/* Number of suffixes marked in the bucket */
	marked = 0;

	/* Search outward from the first suffix,
	 * use binary*/
	while(marked < n && (p >= bucket || q < bucket_end)) {
		if(p >= bucket) {
			/* Scan bucket elements in front */
			if(test_and_mark(sufarray,n,sp,bucket,p,spos)) ++marked;
			p--;
		}

		if(q < bucket_end) {
		/* Scan bucket elements behind */
			if(test_and_mark(sufarray,n,sp,bucket,q,spos)) ++marked;
			q++;
		}
	}
	
	assert(marked >= n);

	if(p < bucket) p++;
	if(q > bucket_end) q--;

	/* Now we scan the bucket left-to-right and derive
	 * the sort order of the input set from the sort order
	 * of the marked suffixes (Step 4-is) */
	while(p <= q) {
#ifdef MULTITHREADED
		if(sp->threaded_sort) {
			assert(p >= bucket && (size_t)(p - bucket) < (sp->marked_suffixes_size << 3));

			if(BITVEC_TEST(sp->marked_suffixes, p - bucket)) {
				*(sufarray++) = (*p) - spos;
			}
		} else {
#endif
			if(*p & MSB_MASK) {
				/* Remove mask to ensure clean results */
				*p ^= MSB_MASK;
				*(sufarray++) = (*p) - spos;
			}
#ifdef MULTITHREADED
		}
#endif

		p++;
	}
}

#define TRY_UPDATE \
{\
	if(*(sp->tp + sufarray[i] + diff)     != current_subbucket[0] ||\
		 *(sp->tp + sufarray[i] + diff + 1) != current_subbucket[1])\
	{\
		/* Try to find the segment where the difference between
			 the given suffix and the anchor suffix is minimal */\
		if(diff < frwd.pos) SETANCHORINFO(frwd,diff,i,segment);\
	}\
	else if(diff < same.pos) SETANCHORINFO(same,diff,i,segment);\
}
 
static size_t split_group(UINT *a, size_t n, UINT lcp, const sort_params_t* sp,
													UINT offset, size_t pivot, size_t *first)
{
	UINT *pb, *pc, *pa_old, *pd_old;
	UINT t;
	size_t r;
	int cmp_result;

	assert(n > 1);

	UINT pivot_pos = a[pivot];
	const fid_Symbol *base = sp->tp + lcp;
	const fid_Symbol *maxptr = base + offset;

	UINT *pa = a;
	UINT *pd = a + n - 1;

	for (/* nothing */; pa != pd && base < maxptr; ++base) {
		pb = pa_old = pa;
		pc = pd_old = pd;

		const fid_Symbol partval = base[pivot_pos];

		for(;;)
		{
			while (pb <= pc &&
						 (cmp_result = cmp_suf_depth(base + *pb, &partval)) <= 0) {
				if (cmp_result == 0) {
					swap(pa, pb);
					pa++;
				}
				pb++;
			}

			while (pb <= pc &&
						 (cmp_result = cmp_suf_depth(base + *pc, &partval)) >= 0) {
				if (cmp_result == 0) {
					swap(pc, pd);
					pd--;
				}
				pc--;
			}

			if (pb > pc) break;

			swap(pb, pc);
			pb++;
			pc--;
		}

		r = min(pa - pa_old, pb - pa);
		vecswap(pa_old, pb - r, r);
		r = min(pd - pc, pd_old - pd);
		vecswap(pb, pd_old + 1 - r, r);
		pa = pa_old + (pb - pa);
		pd = pd_old - (pd - pc);
	}

	*first = pa - a;
	assert(pd >= pa);
	return pd-pa+1;
}

static int backward_anchor_ok(const UINT *sufarray, size_t n,
															const fid_Symbol *seq, Anchorinfo back)
{
	for(size_t i=0; i < n; ++i)
	{
		if(sufarray[i] < back.pos) return -1;
	}

	const fid_Symbol *first=seq+sufarray[0]-back.pos;
	for(size_t i=1; i < n; ++i)
	{
		if(memcmp(first,seq+sufarray[i]-back.pos,back.pos) != 0) return -1;
	}

	return 0;
}

/*!
 * \brief Attempts to sort suffixes using already sorted sub-buckets
 *
 * If the input set of suffixes has a longest common prefix that
 * contains a pair of symbols that matches an already sorted
 * sub-bucket we can use this to sort the suffixes using the
 * sort order of the bucket. If no such symbol pair exists
 * we use the quicksort / blind trie sorting algorithm
 * as a fallback .
 * See Manzini / Ferragina "Engineering a Lightweight Suffix 
 * Array Construction Algorithm" for details.
 * \param sufarray Pointer to the first suffix in the array.
 * \param n Number of suffixes in the array
 * \param lcp Length of the longest common prefix of all
 * 	suffixes within the array.
 * \param sp Sort parameters.
 */
void induced_sort(UINT* sufarray, const size_t n,
		    const UINT lcp, const sort_params_t* sp) {

	/* Sanity checks: */
	if(lcp < 3) {
		/* No sufficiently long common prefix, switch
		 * to quicksort */
		qsort_suf(sufarray, n, sp, lcp);	
		return;
	}

	if(n < 2) {
		/* Do not sort single-element sets */
		return;
	}

	Anchorinfo frwd, same, back;

	SETANCHORINFO(frwd,INVALIDVALUE,n+1,INVALIDVALUE);
	back=same=frwd;

	#ifdef MULTITHREADED
	/* Re-set the thread specific bit vector
		used for marking suffixes */
	if(sp->threaded_sort) {
		memset(sp->marked_suffixes, 0, sp->marked_suffixes_size);
	}
	#endif

	const fid_Symbol current_subbucket[2] =
		{ sp->tp[sufarray[0]], sp->tp[sufarray[0] + 1] };

	UINT segment, diff;
	fid_Uint16 suf_offset;
	size_t i;

	/* Try to find a matching segment as specified by
		 anchor sort */
	for(i = 0; i < n; i++) {
		segment = sufarray[i] / sp->d;
		suf_offset = sufarray[i] % sp->d;
		if(sp->offset[segment] < sp->d)
		{
			if(sp->offset[segment] > suf_offset)
			{
				diff = sp->offset[segment] - suf_offset;
				TRY_UPDATE;
			}
			else
			{
				assert(suf_offset > sp->offset[segment]);
				diff = suf_offset - sp->offset[segment];

				if(diff < back.pos) SETANCHORINFO(back,diff,i,segment);

				if(++segment < sp->seg_count && sp->offset[segment] < sp->d)
				{
					assert(sp->d + sp->offset[segment] > suf_offset);
					diff = sp->d + sp->offset[segment] - suf_offset;
					TRY_UPDATE;
				}
			}
		}
	}

	if(frwd.best != INVALIDVALUE && frwd.pos < lcp-1)
	{
		anchor_sort(sufarray,n,sp,frwd,0);
		return;
	}

	if(back.best != INVALIDVALUE)
	{
		assert(back.pos < sp->d);
		assert(back.pos > 0);
		if(backward_anchor_ok(sufarray,n,sp->tp,back) == 0)
		{
			anchor_sort(sufarray,n,sp,back,1);
			return;
		}
	}

	if(same.best != INVALIDVALUE && same.pos < lcp-1)
	{
		size_t lower;
		size_t equal=split_group(sufarray,n,lcp,sp,same.pos,same.selected,&lower);
		if(equal == n) anchor_sort(sufarray,n,sp,same,0);
		else
		{
			size_t upper=n-equal-lower;
			if(equal > 1) anchor_sort(sufarray+lower,equal,sp,same,0);
			if(lower > 1) qsort_suf(sufarray,lower,sp,lcp);
			if(upper > 1) qsort_suf(sufarray+lower+equal,upper,sp,lcp);
		}
		return;
	}

	/* Could not find a matching anchor, anchor sorting
		 failed, switching to blind trie sort. */
	qsort_suf(sufarray, n, sp, lcp);
	return;
}
