/*
	Copyright (C) 2008, 2009 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#include <libfid.h>

typedef struct
{
	fid_Uintsize uint_size;
	const char *prj_name;
	int update_mode;
} Parameters;

static void copy_arguments(int argc, char *new_argv[], char* argv[])
{
	int dest=2, l1, l2;
	char *opt;
        static char loption[] = "-l";

	new_argv[1] = loption;
	for(int src=1; src < argc; ++src)
	{
		opt=argv[src];
		l1=strncmp(opt,"-w",2);
		l2=strncmp(opt,"--width",7);
		if(l1 != 0 && l2 != 0) new_argv[dest++] = opt;
		else if((l1 == 0 && opt[2] == '\0') || (l2 == 0 && opt[7] == '\0')) ++src;
	}
	new_argv[dest] = NULL;
}

static int parse_launcher_params(int argc, char* argv[], Parameters *params,
																 char *new_argv[]) {
	assert(params != NULL);
	assert(argv != NULL);
	assert(argc > 0);

	copy_arguments(argc,new_argv,argv);

	int r = 0;
	fid_Uint32 temp;

	params->uint_size = fid_UINTSIZE_UNDEF;
	params->update_mode = 1;
	params->prj_name = NULL;

  /* Initialize list of legal command line switches */
  static struct option longopts[] = {
		{"data", required_argument, 0, 'd'},
		{"project", required_argument, 0, 'p'},
		{"width", required_argument, 0, 'w'},
		{0, 0, 0, 0}
  };

	while(r != -1) {
		r = getopt_long(argc, argv, ":d:p:w:", longopts, NULL);

		/* Process each command line switch */
		switch(r) {
			case 'p':
				params->prj_name = optarg;
				break;

			case 'd':
				params->update_mode = 0;
				break;

			case 'w':
				params->uint_size = fid_UINTSIZE_UNDEF;
				if(fid_utils_parse_uint_32(optarg,&temp,NULL) == 0)
				{
					if(temp == 64) params->uint_size = fid_UINTSIZE_64;
				  else if(temp ==  32) params->uint_size = fid_UINTSIZE_32;
				}
				if(params->uint_size == fid_UINTSIZE_UNDEF)
				{
					fprintf(stderr,
									"Error: Invalid width \"%s\", valid options are \"64\" or \"32\".\n"
									"See --help for details.\n", optarg);
					return -1;
				}
				break;

			case ':':
				fprintf(stderr,
								"Error: option %s requires an argument.\n"
								"See --help for details.\n", argv[optind-1]);
				return -1;

			default:	
				break;
		}
	}

	return 0;
}

static int check_or_get_uintsize(const char *prj_name, fid_Uintsize *uisize)
{
	fid_Projectfile prjfile;
	fid_Error error;
	int retcode;

	fid_error_init_default(&error);

	/* Try to read the bit width from the project file in update mode */
	retcode = fid_projectfile_parse_from_file(&prjfile,
																						prj_name, &error);

	if(retcode != 0) {
		/* The project file is considered an integral part of a given project.
		 * Therefore not being able to load it is considered to be a fatal error. */
		fprintf(stderr, "Error while loading project file for update mode.\n");
		fid_error_dump(&error, stderr);
		fprintf(stderr,
						"If you would like to create a new project from scratch, then please specify a\n"
						"data file using the -d switch. See --help for details.\n");
		retcode=-1;
	}
	else
	{
		if(*uisize == fid_UINTSIZE_UNDEF)
		{
			/* Use the bit width from the project file */
			*uisize = prjfile.uisize;
		}
		else if(*uisize != prjfile.uisize) {
			/* Compare the bit width from the project file & from the command line */

			fprintf(stderr,
							"Error: The integer size specified via the -w switch (%u bit) does not match the\n"
							"project's integer size (%u bit). Please pass -w %u (or simply omit the -w\n"
							"switch on the command line) to carry on with the project's integer size in\n"
							"update mode, or re-create the project from scratch with your specified integer\n"
							"size using switch -d. See --help for details.\n",
							(*uisize == fid_UINTSIZE_32) ? 32 : 64,
							prjfile.integer_size_in_bits, prjfile.integer_size_in_bits);
			retcode=-1;
		}
		fid_projectfile_free(&prjfile);
	}

	fid_error_free(&error);
	return retcode;
}

int main(int argc, char* argv[]) {
	Parameters params;
	char* new_argv[argc+2];

	if(parse_launcher_params(argc, argv, &params, new_argv) < 0)
		return EXIT_FAILURE;

	if(params.prj_name == NULL) params.update_mode = 0;

	if(params.update_mode &&
		 check_or_get_uintsize(params.prj_name,&params.uint_size) == -1)
	{
		return EXIT_FAILURE;
	}
	else if(params.uint_size == fid_UINTSIZE_UNDEF)
	{
		/* default to 32 bits */
		params.uint_size = fid_UINTSIZE_32;
	}

	/* try calling executable from argv[0] with "32" or "64" attached */
	size_t len=strlen(argv[0]);
	char executable_with_path[len+3];

	if(len > 0) memcpy(executable_with_path,argv[0],len);
	if(params.uint_size == fid_UINTSIZE_32)
	{
	  executable_with_path[len]='3';
	  executable_with_path[len+1]='2';
	}
	else
	{
	  executable_with_path[len]='6';
	  executable_with_path[len+1]='4';
	}
	executable_with_path[len+2]='\0';
	new_argv[0]=executable_with_path;

	execv(new_argv[0], new_argv);
	int execv_errno=errno;

	/* executable not found via argv[0], try in path */
	static char progname32[] = "mkesa32";
	static char progname64[] = "mkesa64";
	char *progname;

	if(params.uint_size == fid_UINTSIZE_32) progname = progname32;
	else progname = progname64;

	new_argv[0] = progname;
	execvp(new_argv[0], new_argv);
	int execvp_errno=errno;

	/* Goddamnit! */

	fprintf(stderr,
					"Could not launch mkesa. Neither is there an executable named \"%s\"\n"
					"(try 1), nor could program \"%s\" be found in the search path (try 2).\n"
					"Giving up.\n\n", executable_with_path, progname);

	errno=execv_errno;
	perror("Error for try 1 (execv())");
	errno=execvp_errno;
	perror("Error for try 2 (execvp())");

	return EXIT_FAILURE;
}
