/*
	Copyright (C) 2008, 2009 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <libfid.h>
#ifdef MKESA_DATA64
#include <libfid.64>
#else
#include <libfid.32>
#endif

#include "common.h"
#include "algo.h"
#include "cli.h"
#include "fid.h"
#include "qsorts.h"
#include "util.h"

#ifdef MULTITHREADED
#include "threads.h"
#endif

#ifdef MULTITHREADED
static void sort_bucket(sort_params_t *const sp, 
		const mkesa_params_t* params, const fid_Uint16 bucket,
		const pthread_t* threads, const thread_arg_t* thread_args) {

#else
static void sort_bucket(sort_params_t *const sp, 
		const mkesa_params_t* params, const fid_Uint16 bucket) {
#endif

	fid_Uint16 i;
	size_t n;

	const bucket_info_t* buckets = sp->info;

	#ifdef MULTITHREADED
	fid_Uint16 num_jobs = 0;
	if(params->threaded) {
		/* Reset worker thread information */
		*(thread_args->queue_pos) = 0;
		*(thread_args->ready_count) = 0;
	}
	#endif


	for(i = 0; i < buckets->bucket_count; i++) {

		if(i == bucket)
		{
			continue;
		}

		/* If the bucket B(i) is already sorted then 
		 * B(bucket, i) will be sorted as well due to the
		 * copying mechanism. */
		if(IS_SBUCKET_SORTED(buckets,bucket,i)) {
			continue;
		}


		/* Calculate the number of suffixes in the bucket
		 * from the bucket pointers. */
		/* TODO: Make this less ugly. */
		if(i != buckets->bucket_count - 1) {
			n = ((size_t)buckets->sub_buckets[bucket][i + 1]) 
				- ((size_t)buckets->sub_buckets[bucket][i]);
			n /= sizeof(UINT);
		} else {
			n =((size_t)buckets->sub_buckets[bucket][i] - 
				 (size_t)buckets->sub_buckets[bucket][0]);
			n /= sizeof(UINT);

			n = buckets->bucket_sizes[bucket] - n;

		}


		if(i != buckets->bucket_count - 1) {
			#ifdef MULTITHREADED

			if(params->threaded) {
				if(params->verbose > 1) {
					printf("Creating new job for sub-bucket %d/%d, size: %lu\n",
							bucket, i, (unsigned long) n);
				}

				/* Insert new job */
				thread_args->job_queue[num_jobs].bucket[0] = bucket;
				thread_args->job_queue[num_jobs].bucket[1] = i;
				thread_args->job_queue[num_jobs].n = n;
				thread_args->job_queue[num_jobs].lcp = 2;

				/* Announce job to the workers */
				sem_post(thread_args->queue_semaphore);

				num_jobs++;
			} else {
				#endif
			/* Non-threaded mode */
			/* Use multikey quicksort for shallow 
			 * sorting of regular buckets. */
			if(params->verbose > 1) {
				printf("Sorting sub-bucket %d/%d, size: %lu\n",
						bucket, i, (unsigned long) n);
			}

			multikey_qsort_suf(buckets->sub_buckets[bucket][i],
					n, sp, 2);
			SET_SBUCKET_SORTED(buckets,bucket,i);

		#ifdef MULTITHREADED
			}
		#endif
		} else {
			/* If we're dealing with a special-character sub-bucket 
			 * then all omparrisons between suffixes will end after
			 * one character. */
			/* Don't use threading for the special bucket, this will be
			 * the main thread's job while the other threads execute */
			if(params->verbose > 1) {
				printf("Sorting special sub-bucket %d/%d, size: %lu\n",
						bucket, i, (unsigned long) n);
			}

			qsort_offset((void *) buckets->sub_buckets[bucket][i], n);
			SET_SBUCKET_SORTED(buckets,bucket,i);
		}
	}

	#ifdef MULTITHREADED 
	if(params->threaded) {
		if(params->verbose > 1) {
			printf("Waiting for %d job(s) to finish: ", num_jobs);
			fflush(stdout);
		}

		pthread_mutex_lock(thread_args->ready_lock);
		
		while(*(thread_args->ready_count) != num_jobs) {
			/* Wait for all jobs to be finished */
			pthread_cond_wait(thread_args->ready_cond,
					thread_args->ready_lock);

			if(params->verbose > 1) {
				/* Print progress message */
				printf("%d ", *(thread_args->ready_count));
				fflush(stdout);
			}
		}
		
		pthread_mutex_unlock(thread_args->ready_lock);

		if(params->verbose > 1) {
			printf("\nAll jobs finished for bucket %d\n", bucket);
		}
	}
	#endif
}

/*!
 * \brief Sorts the special character bucket.
 *
 * The special character bucket is sorted according to the
 * offset of each suffix.
 * \param buckets Bucket information.
 */
static void sort_special_bucket(bucket_info_t *const buckets) {

	fid_Symbol special = buckets->bucket_count - 1;
	size_t n;

	n = buckets->bucket_sizes[special]; 

	qsort_offset((void *) buckets->sub_buckets[special][0], n);

	/* Mark the entire bucket as sorted */
	for(fid_Uint16 i=0; i < buckets->bucket_count; ++i)
		SET_SBUCKET_SORTED(buckets,special,i);
	SET_WHOLE_BUCKET_SORTED(buckets,special);
}

int main(int argc, char* argv[]) {
	mkesa_params_t params;
	fid_Alphabet alpha;
	fid_Suffixarray esa;
	fid_Uint16 i, b;
	bucket_info_t buckets;

	FILE* suf_file;
	char* filename;
	size_t bytes_read;

	#ifdef MULTITHREADED
	pthread_mutex_t anchor_mutex;
	pthread_t* threads = NULL;
	thread_arg_t thread_args;
	#endif	
	
	/* Parse user (CLI) input */
	if(parse_mkesa_params(argc, argv, &params) < 0) {
		return EXIT_FAILURE;
	} 

	if(check_params_sanity(&params) == -1) {
		return EXIT_FAILURE;
	}

	/* Generate / load and store the alphabet */
	if(get_alphabet(&alpha, &params) != 0) {
		return EXIT_FAILURE;
	}

	/* Transform the input sequence, build various tables */
	if(build_tables(&alpha, &esa, &params) != 0) {
		return EXIT_FAILURE;
	}

#ifndef MKESA_DATA64
	/* Quick check to avoid overrunning 32 bit data structures. */
	if(esa.sequences.total_length.v_uint32 > ((fid_Uint32) 0x7FFFFFF0))  {
		fprintf(stderr, "Error: Input sequence is too long for 32 bit data\n\
structures, try using 64 bit data structures.\n(See --help for details).\n");
		return EXIT_FAILURE;
	}
#endif

	if(!params.update_mode || 
		!test_file_readable(params.prj_name, "suf")) {
		/* Generate suffix array from scratch
		since we're either in create mode or the suffix
		array doesn't exist yet. */
		
		bucket_info_init(&buckets);

		if(params.verbose) {
			printf("\nCommencing initial sort to buckets: ");
			fflush(stdout);
		}

		/* Sort the suffixes into buckets,
		 * also generates bucket information */
		buckets_tis_to_suf(&esa, &buckets);
	
		if(params.verbose) {
			printf("sort complete.\n");
		}

		if(params.verbose > 1) {
			printf("Largest sub-bucket size: %lu\n", 
				(unsigned long) buckets.largest_sbucket_size);
		}

		/* Initialize pointers used by the sorting algorithms */
		fid_Symbol* text_pos = esa.sequences.tisfile.content;
		fid_Symbol* text_end =
			esa.sequences.tisfile.content + esa.sequences.total_length.VU;


		/* Initialize structure with sort parameters */
		sort_params_t sp;

		/* Input text boundaries */
		sp.tp = text_pos;
		sp.tm = text_end;
		sp.info = &buckets;
		sp.maxdepth = params.maxdepth;
		sp.blind_threshold = esa.sequences.total_length.VU / 2000;

		sp.largest_sbucket_size = buckets.largest_sbucket_size;
		sp.marked_suffixes = NULL;
		sp.marked_suffixes_size = BITVEC_BYTES(sp.largest_sbucket_size);


		/* Allocate memory for anchor sorting */
		sp.d = params.segment_size;
		sp.seg_count = (esa.sequences.total_length.VU / sp.d) + 1;
		
		/* Allocate memory for the offset and anchor arrays */
		sp.offset = malloc_safe(sizeof(fid_Uint16) * sp.seg_count);
		sp.anchor = malloc_safe(sizeof(UINT*) * sp.seg_count);
	
		assert(sp.offset != NULL);
		assert(sp.anchor != NULL);
	
		for(size_t j = 0; j < sp.seg_count; ++j)
		{
			sp.offset[j] = sp.d;
			sp.anchor[j] = NULL;
		}
	
		#ifdef MULTITHREADED
		sp.threaded_sort = params.threaded;
			
		if(params.threaded) {
			if(params.thread_count > buckets.bucket_count - 1) {
				/* Do not spawn more threads than we can actually use */
				printf(
"Warning: Reducing thread count from %d to %d to avoid idle threads\n",
				params.thread_count, buckets.bucket_count - 1);
				params.thread_count = buckets.bucket_count - 1;
			}

		/* Initialize the anchor mutex */
			pthread_mutex_init(&anchor_mutex, NULL);
			sp.anchor_mutex = &anchor_mutex;

			if(params.verbose) {
				printf("Starting %d worker threads.\n",
						params.thread_count);	
			}
		
			/* Start the worker threads */
			threads = start_threads(&thread_args,
					params.thread_count, &sp);
			assert(threads != NULL);
		}
		#endif
	
		if(params.verbose) {
			printf("\nSorting special bucket `%c' (size %lu).\n", 
					fid_PRINT_SYMBOL(&alpha,buckets.bucket_count - 1),
					(unsigned long) buckets.bucket_sizes[buckets.bucket_count - 1]);
		}
	
		/* Sort the special character bucket */
		sort_special_bucket(&buckets);
	
		/* Start the sorting process */
		if(params.verbose > 2) {
			printf("\nSorting buckets: (Parameters: L="UINTFMT", d=%hu)\n",
				sp.maxdepth, sp.d);
		}
	
		for(i = 0; i < buckets.bucket_count - 1; i++) {
			/* Sort each normal bucket according to their size.
			 * (The smallest bucket is sorted first to increase 
			 * copy sorting effectivity) */
			b = buckets.sort_order[i];
	
			if(params.verbose) {
				printf("\nSorting bucket `%c' (size %lu).\n",
						fid_PRINT_SYMBOL(&alpha,b),
						(unsigned long) buckets.bucket_sizes[b]);
			}
	
			/* Sort the bucket */
			#ifdef MULTITHREADED
			sort_bucket(&sp, &params, b, threads, &thread_args);
			#else
			sort_bucket(&sp, &params, b);
			#endif
	
			if(params.verbose > 1) {
				printf("Sorting related sub-buckets.\n");
			}

			copy_sort(&esa, &buckets, b);
	
			/* Mark the entire bucket as sorted */
			SET_WHOLE_BUCKET_SORTED(&buckets,b);
		}
	
		#ifdef MULTITHREADED
		/* Destroy the anchor mutex */
		if(params.threaded) {
			pthread_mutex_destroy(&anchor_mutex);
	
			if(params.verbose) {
				printf("\nShutting down worker threads.\n");
			}
	
			/* Shut down the worker threads */
			kill_threads(threads, &thread_args,
					params.thread_count);
		}
		#endif

		/* Write the results */
		write_and_free_tables(&esa, &params);

		/* Clean-up */
		bucket_info_free(&buckets);

		/* Clean up the memory used by anchor sorting */
		free(sp.offset);
		free(sp.anchor);

	} else if(params.build_lcp || params.build_sti) {
		/* Only load the suffix array from the
			 existing file if we actually need it */

		if(params.verbose) {
			printf("\nLoading existing suffix array file.\n");
		}

		/* Generate file name */
		filename = fid_filename_create(params.prj_name, "suf", NULL);

		suf_file = fopen(filename, "r");

		if(suf_file == NULL) {
			perror("Couldn't open suffix array file for reading even\
though it supposedly exists (internal error).\n");
			return EXIT_FAILURE;
		}
	
		/* Read data from the actual suffix array file on the disc
			 into the pre-allocated fake file in memory */
		bytes_read = fread(esa.suffile.content, sizeof(UINT),
				esa.sequences.total_length.VU + 1, suf_file);

		if(bytes_read != esa.sequences.total_length.VU + 1) {
			/* TODO: Change this to use stat() */
			fprintf(stderr, "Fatal error: Could not read the required\
data from the suffix array\nfile \"%s\" (Read: %lu, expected \
%lu symbols)\n", filename, (unsigned long) bytes_read,
					(unsigned long) esa.sequences.total_length.VU + 1);
			return EXIT_FAILURE;
		}

		fclose(suf_file);
		free(filename);

		params.write_suf = 0;

		/* Write the results */
		write_and_free_tables(&esa, &params);

	} else {
		if(params.verbose) {
			printf("\nNothing to be done in update mode, exiting.\n");
		}
	}

	fid_memory_show_usage(stdout);

	return EXIT_SUCCESS;
}

