/*
	Copyright (C) 2008 David Fleer, Robert Homann.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef MULTITHREADED
#include <pthread.h>
#endif

#include "common.h"
#include "algo.h"
#include "blind.h"
#include "induced.h"
#include "qsorts.h" 

/*!
 * \brief Insertion sort for suffix arrays.
 *
 * \param a Pointer to the suffix array.
 * \param n Number of elements in the suffix array.
 * \param lcp Length of the longest common prefix
 *	of all the suffixes within the suffix array.
 * \param sp Sort parameters.
*/
static inline void isort(UINT *const a, const size_t n,
		const UINT lcp, const sort_params_t* sp) {
	UINT *pm, *pl, t;
	const fid_Symbol* base = sp->tp + lcp;

	/* Find the next element that is out of order. */
	for(pm = a + 1; pm < a + n; pm++) {
		/* Move the element until it is in order. */
		for(pl = pm; pl > a && \
			cmp_suf_lcp_ro(base + *(pl - 1), base + *pl) > 0; pl--) {
				/* Move the element */
				swap(pl, pl - 1);
		}
	}
}

/* Insertion sort macro for use within the offset quicksort.*/
#define ISORT_OFFSET {\
	for (pm = a + 1; pm < a + n; pm++) {\
		for (pl = pm; pl > a && *(pl - 1) > *pl; pl--) {\
			swap(pl, pl - 1);\
		}\
	}\
}

#define ISORT_RANK {\
	for (pm = a + 1; pm < a + n; pm++) {\
		for (pl = pm; pl > a && (sufarray[*(pl - 1)] > sufarray[*pl]); pl--) {\
			swap(pl, pl - 1);\
		}\
	}\
}


/*!
 * \brief Calculates the pseudomedian of a suffix array.
 * 
 * If the number of suffixes is smaller than 30, then a mere
 * pseudomedian of three is used. Otherwise a median of
 * three medians of three is used calculate the pseudomedian.
 *
 * The \p med3_func is used to find the median of three.
 *
 * \param a Beginning of the suffix array.
 * \param n Number of suffixes in the suffix array.
 * \param lcp Lenght of the longest common prefix of all
 * 	suffixes in the array.
 * \param med3_func The function used to determine the
 * 	median of three.
 * \param sp Structure with various sorting parameters.
 * \returns Pointer to the index of the median.
 */
static inline UINT* pseudomedian(const UINT* a, const size_t n, const UINT lcp,
		const sort_params_t* sp, const UINT* (*med3_func)(const UINT*, const UINT*,
		const UINT*, const fid_Symbol*, const UINT)) {

	/* Chose three elements from the beginning, middle
		 and end of the suffix array. */
	const UINT* pl = a;
	const UINT* pm = a + (n / 2);
	const UINT* pn = a + (n - 1);

	if (n > 30) {
		const size_t d = (n / 8);
		/* Three medians of three */
		pl = (*med3_func)(pl, pl + d, pl + 2 * d, sp->tp, lcp);
		pm = (*med3_func)(pm - d, pm, pm + d, sp->tp, lcp);
		pn = (*med3_func)(pn - 2 * d, pn - d, pn, sp->tp, lcp);
	}
	/* Median of three of the previous results */
	return (UINT*) (*med3_func)(pl, pm, pn, sp->tp, lcp);
}

/*!
 * \brief Calculates the pseudomedian of three/nine of a suffix array sorted
 * by the offset of each suffix.
 *
 * \see pseudomedian
 *
 * \param a Beginning of the suffix array.
 * \param n Number of suffixes in the suffix array.
 * \returns Pointer to the index of the median.
 */
static inline UINT* pseudomedian_offset(const UINT* a, const size_t n) {

	/* Chose three elements from the beginning, middle
		 and end of the suffix array. */
	const UINT* pl = a;
	const UINT* pm = a + (n / 2);
	const UINT* pn = a + (n - 1);

	if (n > 30) {
		const size_t d = (n / 8);
		/* Three medians of three */
		pl = med3_offset(pl, pl + d, pl + 2 * d);
		pm = med3_offset(pm - d, pm, pm + d);
		pn = med3_offset(pn - 2 * d, pn - d, pn);
	}

	/* Median of three of the previous results */
		return (UINT*) med3_offset(pl, pm, pn);
}

static inline UINT* pseudomedian_rank(const UINT* a, const size_t n,
		const UINT* sufarray) {

	/* Chose three elements from the beginning, middle
		 and end of the suffix array. */
	const UINT* pl = a;
	const UINT* pm = a + (n / 2);
	const UINT* pn = a + (n - 1);

	if (n > 30) {
		const size_t d = (n / 8);
		/* Three medians of three */
		pl = med3_rank(pl, pl + d, pl + 2 * d, sufarray);
		pm = med3_rank(pm - d, pm, pm + d, sufarray);
		pn = med3_rank(pn - 2 * d, pn - d, pn, sufarray);
	}

	/* Median of three of the previous results */
		return (UINT*) med3_rank(pl, pm, pn, sufarray);
}


/*!
 * \brief Multikey quicksort for a set of suffixes.
 *
 * Sorts suffixes in \p a using multikey quicksort with
 * optional ternary quicksort / blind trie sorting.
 * A pivot is chosen by using pseudomedian of three/nine.
 * The suffixes are then sorted by comparing a single symbol.
 * Those suffixes that are lesser or greater than the pivot are
 * sorted by multikey quicksort using the same depth,
 * those that were equal to the pivot are sorted while increasing
 * the depth by one.
 * If the sorting depth is greater than the maximum depth specified
 * in the sort_params_t structure, the sorting algorithm
 * is switched to induced/anchor sort.
 *
 * \see qsort_suf
 * \see induced_sort
 * \see anchor_sort
 * \param a Pointer to the first suffix in the set.
 * \param n Number of suffixes in the set.
 * \param sp Sort parameters.
 * \param depth Current comparison depth.
 */
void multikey_qsort_suf(UINT *a, size_t n, const sort_params_t* sp, 
		UINT depth)
{
	/* See Bentley/McIlroy for the actual meaning of these pointers. */
	UINT *pa, *pb, *pc, *pd, *pm, *pn;
	const fid_Symbol *base = sp->tp + depth;
	UINT t;
	size_t r;
	int cmp_result;
	int swap_cnt;

	if(n <= 1) return;

multi_loop:	
	if(depth >= sp->maxdepth) {
		/* Switch to the deep sorting algorithm once we've
		  passed the depth threshold */
		/* Use induced sort as our initial deep-sorting algorithm */
		induced_sort(a, n, depth, sp);
		update_anchor(a, n, sp);
		
		return;
	}
	
	/* Sort small sets using insertion sort */
	if (n < 20) {
		/* ISORT(depth); */
		isort(a, n, depth, sp);
		return;
	}
	
	/* Chose the pivot by a pseudomedian of 3 or 9 */
	pm = pseudomedian(a, n, depth, sp, &med3_depth);

	/* Move the pivot to the beginning of the array. */
	swap(a, pm);

	const fid_Symbol *pivot = base + *a;

	/* Set the a and b pointers to the first element after the pivot. */
	pa = pb = a + 1;

	/* Set the c and d pointers to the last element. */
	pc = pd = a + (n - 1);

	cmp_result = 0;
	swap_cnt = 0;
	
	for (;;) {
		/* Increment the b pointer over all the elements at the beginning of
		 * the array until we encounter an element that's larger than the
		 * pivot. */

		while (pb <= pc && (cmp_result = cmp_suf_depth(base + *pb, pivot)) <= 0) {
			/* Swap all those elements that are equal to the pivot to the
			  beginning of the array. */
			if (cmp_result == 0) {
				swap_cnt = 1;
				swap(pa, pb);
				pa++;
			}
			pb++;
		}

		/* Decrement the c pointer over all the elements at the end of the
		 * array until we encounter an element that's smaller than the
		 * pivot. */

		while (pb <= pc && (cmp_result = cmp_suf_depth(base + *pc, pivot)) >= 0) {
			/* Again, elements that are equal to the pivot are swapped
			  to the end of the array */
			if (cmp_result == 0) {
				swap_cnt = 1;
				swap(pc, pd);
				pd--;
			}
			pc--;
		}

		/* At this point pb points to the first element in the array that
		 * is larger than the pivot while pc points to the last element that
		 * is smaller than the pivot.
		 */

		/* If pb and pc have crossed we're done. */
		if (pb > pc)
			break;

		/* Swap the first larger element with the last smaller element */
		swap(pb, pc);
		swap_cnt = 1;
		pb++;
		pc--;
	}
	if (swap_cnt == 0) {  /* Switch to insertion sort */
		/*ISORT(depth);*/
		isort(a, n, depth, sp);
		return;
	}

	if(pa > pd)
	{
		/* Iterate rather than recurse to save stack space */
		++depth;
		++base;
		goto multi_loop;
	}

	/* Move the blocks of elements that are equal (to the pivot) from
	  the beginning / end to the middle. */
	pn = a + n;
	/* Swap the area at the beginning to the middle. */
	r = min(pa - a, pb - pa);
	vecswap(a, pb - r, r);
	/* Swap the area at the end to the middle */
	r = min(pd - pc, (pn - pd) - 1);
	vecswap(pb, pn - r, r);

	/* If there's more than one element in the "smaller than pivot"
	  block then start sorting that block. */
	if((r = pb - pa) > 1) {
		multikey_qsort_suf(a, r, sp, depth);
	}

	/* If there's more than one element in the "equal to pivot block"
	  then sort that block while increasing the depth. */
	if((r = (pa - a) + ((pn - pd) - 1)) > 1) { 
		multikey_qsort_suf(a + (pb - pa), r, sp, depth + 1);
	}

	/* If there's more than one element in the "larger than pivot"
	  block, then start sorting that block. */
	if((r = pd - pc) > 1) {
		multikey_qsort_suf(pn - r, r, sp, depth);
	}
}

/*!
 * \brief Ternary quicksort for a set of suffixes.
 *
 * Sorts suffixes in \p a using ternary quicksort while switching
 * to blind trie sorting once a certain size threshold is reached.
 * A pivot is chosen by using pseudomedian of three/nine.
 * If the number of suffixes being sorted is less than
 * the blind_threshold specified in the sort_params_t data
 * structure then the sorting algorithm is switched to blind sorting.
 *
 * \see multikey_qsort_suf
 * \see blind_sort
 * \param a Pointer to the first suffix in the set.
 * \param n Number of suffixes in the set.
 * \param sp Sort parameters.
 * \param lcp Length of the longest common prefix of all the suffixes. 
 */
void qsort_suf(UINT *a, size_t n, const sort_params_t* sp, UINT lcp)
{
	/* See Bentley/McIlroy for the actual meaning of these pointers. */
	UINT *pa, *pb, *pc, *pd, *pm, *pn;
	UINT t;

	size_t r;
	int cmp_result;
	int swap_cnt;
	fid_Symbol *base;

	/* The LCP value for those elements that are smaller than the
	  current pivot */
	UINT slcp;
	/* LCP value for elements that are larger than the pivot */
	UINT llcp; 
	/* Temporary LCP value */
	UINT tlcp;

loop:	
	cmp_result = 0;
	swap_cnt = 0;
	base = sp->tp + lcp;

	if(n < sp->blind_threshold) {
		/* Switch to blind sort
		  Reduce LCP by one to avoid out-of-bounds error */
		blind_sort(a, n, sp, (lcp > 0? lcp - 1 : 0));
		return;
	} else if(n < 20) {
	/* Sort small sets using insertion sort */
		/*ISORT(lcp);*/
		isort(a, n, lcp, sp);
		return;
	}

	slcp = ~((UINT) 0);
	llcp = ~((UINT) 0);
	tlcp = lcp;

	/* Chose the pivot by a pseudomedian of 3 or 9 */
	pm = pseudomedian(a, n, tlcp, sp, &med3);


	/* Move the pivot to the beginning of the array. */
	swap(a, pm);

	/* Calculate the pivot symbol */
	const fid_Symbol *pivot = base + *a;

	/* Set the a and b pointers to the first element after the pivot. */
	pa = pb = a + 1;

	/* Set the c and d pointers to the last element. */
	pc = pd = a + (n - 1);

	for (;;) {
		/* Increment the b pointer over all the elements at the beginning of
		 * the array until we encounter an element that's larger than the
		 * pivot. */
		tlcp = lcp;

		while (pb <= pc && (cmp_result = cmp_suf_lcp(
					base + *pb, pivot, &tlcp)) <= 0) {
			/* If the element compared above turns out to be smaller than
			  the pivot: Update the LCP value for small elements. */
			slcp = min(slcp, tlcp);

			/* Reset temporary LCP */
			tlcp = lcp;
			pb++;
		}

		/* Reaching this point might indicate that the while loop encountered
		 * an element that was larger than the pivot, in that case update the
		 * LCP value for large elements. */
		if(cmp_result > 0) {
			llcp = min(llcp, tlcp);
		}
		
		/* Decrement the c pointer over all the elements at the end of the
		 * array until we encounter an element that's smaller than the
		 * pivot. */

		tlcp = lcp;
		while (pb <= pc && (cmp_result = cmp_suf_lcp(
					base + *pc, pivot, &tlcp)) >= 0) {
			/* Update LCP for larger elements */
			llcp = min(llcp, tlcp);
			tlcp = lcp;

			pc--;
		}

		if(cmp_result < 0) {
			slcp = min(slcp, tlcp);
		}

		/* At this point pb points to the first element in the array that
		 * is larger than the pivot while pc points to the last element that
		 * is smaller than the pivot.
		 */

		/* If pb and pc have crossed we're done. */
		if (pb > pc)
			break;

		/* Swap the first larger element with the last smaller element */
		swap(pb, pc);
		swap_cnt = 1;
		pb++;
		pc--;
	}
	if (swap_cnt == 0) {  /* Switch to insertion sort */
		/* ISORT(lcp);*/
		isort(a, n, lcp, sp);
		return;
	}

	/* Move the pivot to the middle of the array */
	pn = a + n;
	
	swap(a, pb - 1);

	/* If there's more than one element in the "smaller than pivot"
	  block then start sorting that block. */
	if ((r = pb - pa) > 1)
		qsort_suf(a, r, sp, slcp);

	/* If there's more than one element in the "larger than pivot"
	  block, then start sorting that block. */
	if ((r = pd - pc) > 1) {
		/* Iterate rather than recurse to save stack space */
		a = pn - r;
		n = r;
		lcp = llcp;
		goto loop;
	}
}

/*!
 * \brief Quicksort using offsets.
 *
 * Sorts suffixes according to their offsets, which is the
 * position of the beginning of the suffix in the input sequence.
 * \param a Pointer to the first suffix in the set.
 * \param n Number of suffixes in the set.
 */
void qsort_offset(UINT *a, size_t n)
{
	/* See Bentley/McIlroy for the actual meaning of these pointers. */
	UINT *pa, *pb, *pc, *pd, *pl, *pm, *pn;
	UINT t;

	size_t r;
	int swap_cnt;
loop:	
	swap_cnt = 0;

	if(n < 20) {
		ISORT_OFFSET;
		return;
	}

	pm = pseudomedian_offset(a, n);
	/* Move the pivot to the beginning of the array. */
	swap(a, pm);

	/* Set the a and b pointers to the first element after the pivot. */
	pa = pb = a + 1;

	/* Set the c and d pointers to the last element. */
	pc = pd = a + (n - 1);

	for (;;) {
		/* Increment the b pointer over all the elements at the beginning of
		 * the array until we encounter an element that's larger than the
		 * pivot. */

		while (pb <= pc && *pb <= *a) {
			pb++;
		}

		/* Decrement the c pointer over all the elements at the end of the
		 * array until we encounter an element that's smaller than the
		 * pivot. */
		while (pb <= pc && *pc >= *a) {
			pc--;
		}

		/* At this point pb points to the first element in the array that
		 * is larger than the pivot while pc points to the last element that
		 * is smaller than the pivot.
		 */

		/* If pb and pc have crossed we're done. */
		if (pb > pc)
			break;

		/* Swap the first larger element with the last smaller element */
		swap(pb, pc);
		swap_cnt = 1;
		pb++;
		pc--;
	}
	if (swap_cnt == 0) {  /* Switch to insertion sort */
		ISORT_OFFSET;
		return;
	}

	/* Move the pivot to the middle of the array */
	pn = a + n;
	
	swap(a, pb - 1);

	/* If there's more than one element in the "smaller than pivot"
	  block then start sorting that block. */
	if ((r = pb - pa) > 1)
		qsort_offset(a, r);

	/* If there's more than one element in the "larger than pivot"
	  block, then start sorting that block. */
	if ((r = pd - pc) > 1) {
		/* Iterate rather than recurse to save stack space */
		a = pn - r;
		n = r;
		goto loop;
	}
}

void qsort_rank(UINT *a, size_t n, const UINT* sufarray) {
	/* See Bentley/McIlroy for the actual meaning of these pointers. */
	UINT *pa, *pb, *pc, *pd, *pl, *pm, *pn;
	UINT t;

	size_t r;
	int swap_cnt;
loop:	
	swap_cnt = 0;

	if(n < 20) {
		ISORT_RANK;
		return;
	}

	pm = pseudomedian_rank(a, n, sufarray);
	/* Move the pivot to the beginning of the array. */
	swap(a, pm);

	/* Set the a and b pointers to the first element after the pivot. */
	pa = pb = a + 1;

	/* Set the c and d pointers to the last element. */
	pc = pd = a + (n - 1);

	for (;;) {
		/* Increment the b pointer over all the elements at the beginning of
		 * the array until we encounter an element that's larger than the
		 * pivot. */

		while (pb <= pc && sufarray[*pb] <= sufarray[*a]) {
			pb++;
		}

		/* Decrement the c pointer over all the elements at the end of the
		 * array until we encounter an element that's smaller than the
		 * pivot. */
		while (pb <= pc && sufarray[*pc] >= sufarray[*a]) {
			pc--;
		}

		/* At this point pb points to the first element in the array that
		 * is larger than the pivot while pc points to the last element that
		 * is smaller than the pivot.
		 */

		/* If pb and pc have crossed we're done. */
		if (pb > pc)
			break;

		/* Swap the first larger element with the last smaller element */
		swap(pb, pc);
		swap_cnt = 1;
		pb++;
		pc--;
	}
	if (swap_cnt == 0) {  /* Switch to insertion sort */
		ISORT_RANK;
		return;
	}

	/* Move the pivot to the middle of the array */
	pn = a + n;
	
	swap(a, pb - 1);

	/* If there's more than one element in the "smaller than pivot"
	  block then start sorting that block. */
	if ((r = pb - pa) > 1)
		qsort_rank(a, r, sufarray);

	/* If there's more than one element in the "larger than pivot"
	  block, then start sorting that block. */
	if ((r = pd - pc) > 1) {
		/* Iterate rather than recurse to save stack space */
		a = pn - r;
		n = r;
		goto loop;
	}
}

