/*
	Copyright (C) 2008, 2009 David Fleer.

	This file is part of mkesa.

	mkesa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	mkesa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with mkesa.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef MULTITHREADED

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_SYSCONF
#include <unistd.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "common.h"
#include "algo.h"
#include "threads.h"
#include "qsorts.h"
#include "util.h"

void* start_sort_thread(void* options) {
  thread_arg_t* args = (thread_arg_t*) options;
	sort_job_t* job;

	/* Create a thread-specific copy of the sort
	parameter. This is needed in order to pass the 
	thread-specific bit vectors to the sort algorithms. */
	sort_params_t *local_sp;
	local_sp = malloc_safe(sizeof(sort_params_t));
	memcpy(local_sp, args->sp, sizeof(sort_params_t));

	/* Make sure that the local copy is cleaned up upon
	thread exit */
	pthread_cleanup_push(free, local_sp);

	/* Allocate memory for thread-local marking 
	of suffixes during the induced sort */
	local_sp->marked_suffixes = malloc_safe(
		local_sp->marked_suffixes_size);

	pthread_cleanup_push(free, local_sp->marked_suffixes);

	while(1) {
		/* Thread main loop */
	
		/* Cancel immediately while waiting */
		pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,
			NULL);

		/* Wait for new jobs from the main thread */
		sem_wait(args->queue_semaphore);

		/* Do not cancel while we're working on a job */
		pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,
			NULL);

		/* Lock the job queue and get a new job */
		pthread_mutex_lock(args->queue_lock);


		/* Acquire job, increase queue position */
		job = &(args->job_queue[*(args->queue_pos)]);
		*(args->queue_pos) = (*(args->queue_pos) + 1)
											% args->queue_size;
		
		/* Release the queue lock */
		pthread_mutex_unlock(args->queue_lock);

		/* Do the actual work */
 		multikey_qsort_suf(local_sp->info->sub_buckets[job->bucket[0]][job->bucket[1]],
					job->n, local_sp, job->lcp);

		SET_SBUCKET_SORTED(local_sp->info,job->bucket[0],job->bucket[1]);

		/* Lock the ready count mutex */
		pthread_mutex_lock(args->ready_lock);
		/* Increase the ready count to let the parent
			 know that we've finished another job */
		(*(args->ready_count))++;
		/* Release the ready count mutex */
		pthread_mutex_unlock(args->ready_lock);
		
		/* Wake up the parent */
		pthread_cond_signal(args->ready_cond);

		/* Test for cancellation requests */
		pthread_testcancel();
	}

	/* Execute cleanup */
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);

	return NULL;
}

pthread_t* start_threads(thread_arg_t* thread_args,
		const fid_Uint16 thread_count, sort_params_t* sp) {
	fid_Uint16 i;

	assert(thread_args != NULL);
	assert(sp != NULL);
		
	/* Copy sort parameters */
	thread_args->sp = sp;

	/* Initialize job queue memory */
	thread_args->job_queue = malloc_safe(sizeof(sort_job_t) * 
			(sp->info->bucket_count - 1));
	assert(thread_args->job_queue != NULL);
	
	/* Initialize the mutex that governs access
		 to the job queue. */
  thread_args->queue_lock = malloc_safe(sizeof(pthread_mutex_t));
	assert(thread_args->queue_lock != NULL);
	pthread_mutex_init(thread_args->queue_lock, NULL);

	/* Initialize the semaphore that manages the job
		 queue and informs the threads about the number
		 of jobs available. */
  thread_args->queue_semaphore = malloc_safe(sizeof(sem_t));
	assert(thread_args->queue_semaphore != NULL);
	sem_init(thread_args->queue_semaphore, 0, 0);

	/* Initialize queue position and ready count, where
		 queue_pos is the position of the next available job
		 within the queue. Ready count is incremented every
		 time a job is completed by one of the worker threads. */
  thread_args->queue_pos = malloc_safe(sizeof(fid_Uint16));
  thread_args->ready_count = malloc_safe(sizeof(fid_Uint16));
	assert(thread_args->queue_pos != NULL);
	assert(thread_args->ready_count != NULL);

  *(thread_args->queue_pos) = 0;
  *(thread_args->ready_count) = 0;

	/* Queue size = alphabet size - 1 since that's the maximum
		 number of jobs that can be available at the same tim */
  thread_args->queue_size = (sp->info->bucket_count - 1);

	/* Initialize the mutex that governs access to the ready_count
		 condition variable. */
  thread_args->ready_lock = malloc_safe(sizeof(pthread_mutex_t));
	assert(thread_args->ready_lock != NULL);
	pthread_mutex_init(thread_args->ready_lock, NULL);

	/* Initialize the condition structure used for the ready_count
		 condition variable. */
  thread_args->ready_cond = malloc_safe(sizeof(pthread_cond_t));
	assert(thread_args->ready_cond != NULL);
	pthread_cond_init(thread_args->ready_cond, NULL);

	/* Allocate memory for the thread structure structure. */
	pthread_t* threads = malloc_safe(sizeof(pthread_t) * thread_count);
	assert(threads != NULL);

	/* Create worker threads. */
	for(i = 0; i < thread_count; i++) {
		if(pthread_create(threads + i, NULL, &start_sort_thread, 
				thread_args)) {
			fprintf(stderr,
					"Critical error: Could not create thread.\n");

			#ifdef HAVE_ERRNO_H
			perror(NULL);
			#endif

			/* If we can't create new threads that indicates a
      * major fubar, abort */
      exit(1);
    }
	}

	return threads;
}

void kill_threads(pthread_t* threads, 
		thread_arg_t* thread_args, const fid_Uint16 thread_count) {

	fid_Uint16 i;
	for(i = 0; i < thread_count; i++) {
		/* Kill and join all threads */
		pthread_cancel(threads[i]);
		pthread_join(threads[i], NULL);
	}

	/* Clean up the job queue */
	free(thread_args->job_queue);

	/* Clean up the mutexes */
	pthread_mutex_destroy(thread_args->queue_lock);
	free(thread_args->queue_lock);

	pthread_mutex_destroy(thread_args->ready_lock);
	free(thread_args->ready_lock);

	/* Clean up the semaphore */
	sem_destroy(thread_args->queue_semaphore);
	free(thread_args->queue_semaphore);

	/* Clean up the condition variable */
	pthread_cond_destroy(thread_args->ready_cond);
	free(thread_args->ready_cond);

	free(thread_args->queue_pos);
	free(thread_args->ready_count);
	
	/* Free thread memory */
	free(threads);
}

#ifdef HAVE_SYSCONF
unsigned int get_num_of_cpus(void) {
	long count=0;

#ifdef _SC_NPROCESSORS_ONLN
	if((count=sysconf(_SC_NPROCESSORS_ONLN)) < 1) {
#endif /* _SC_NPROCESSORS_ONLN */

#ifdef _SC_NPROCESSORS_CONF
		count=sysconf(_SC_NPROCESSORS_CONF);
#endif /* _SC_NPROCESSORS_CONF */
		if(count < 1) {
			count=0;
		}
#ifdef _SC_NPROCESSORS_ONLN
	}
#endif /* _SC_NPROCESSORS_ONLN */
	return (unsigned int)count;
}
#endif

#endif
